@@ -187,6 +187,10 @@ static cl::opt<bool, true> unittest("unittest",
                                     cl::desc("Compile in unit tests"),
                                     cl::location(global.params.useUnitTests));
 
+cl::opt<std::string>
+    ir2objCacheDir("ir2obj-cache", cl::desc("Use <cache dir> to cache object files for whole IR modules (experimental)"),
+            cl::value_desc("cache dir"), cl::Prefix);
+
 static StringsAdapter strImpPathStore("J", global.params.fileImppath);
 static cl::list<std::string, StringsAdapter>
     stringImportPaths("J", cl::desc("Where to look for string imports"),
@@ -235,6 +239,11 @@ cl::list<std::string> versions(
     cl::desc("Compile in version code >= <level> or identified by <idents>"),
     cl::value_desc("level/idents"), cl::CommaSeparated);
 
+cl::list<std::string> transitions(
+    "transition",
+    cl::desc("help with language change identified by <idents>, use ? for list"),
+    cl::value_desc("idents"), cl::CommaSeparated);
+
 static StringsAdapter linkSwitchStore("L", global.params.linkswitches);
 static cl::list<std::string, StringsAdapter>
     linkerSwitches("L", cl::desc("Pass <linkerflag> to the linker"),
@@ -274,10 +283,14 @@ cl::opt<std::string>
 
 cl::opt<llvm::Reloc::Model> mRelocModel(
     "relocation-model", cl::desc("Relocation model"),
+#if LDC_LLVM_VER < 309
     cl::init(llvm::Reloc::Default),
+#endif
     cl::values(
+#if LDC_LLVM_VER < 309
         clEnumValN(llvm::Reloc::Default, "default",
                    "Target default relocation model"),
+#endif
         clEnumValN(llvm::Reloc::Static, "static", "Non-relocatable code"),
         clEnumValN(llvm::Reloc::PIC_, "pic",
                    "Fully relocatable, position independent code"),
@@ -366,6 +379,11 @@ cl::opt<bool, true>
     singleObj("singleobj", cl::desc("Create only a single output object file"),
               cl::location(global.params.singleObj));
 
+cl::opt<uint32_t, true> hashThreshold(
+    "hash-threshold",
+    cl::desc("hash symbol names longer than this threshold (experimental)"),
+    cl::location(global.params.hashThreshold), cl::init(0));
+
 cl::opt<bool> linkonceTemplates(
     "linkonce-templates",
     cl::desc(
@@ -410,6 +428,19 @@ cl::opt<unsigned char, true, CoverageParser> coverageAnalysis(
                     "minimum required coverage)"),
     cl::location(global.params.covPercent), cl::ValueOptional, cl::init(127));
 
+#if LDC_WITH_PGO
+cl::opt<std::string>
+    genfileInstrProf("fprofile-instr-generate", cl::value_desc("filename"),
+                     cl::desc("Generate instrumented code to collect "
+                              "execution counts (e.g. for PGO)"),
+                     cl::ValueOptional);
+
+cl::opt<std::string> usefileInstrProf(
+    "fprofile-instr-use", cl::value_desc("filename"),
+    cl::desc("Use instrumentation data for profile-guided optimization"),
+    cl::ValueRequired);
+#endif
+
 static cl::extrahelp footer(
     "\n"
     "-d-debug can also be specified without options, in which case it enables "
