@@ -19,11 +19,16 @@
 #include "gen/optimizer.h"
 #include "gen/programs.h"
 #include "llvm/ADT/Triple.h"
+#include "llvm/IRReader/IRReader.h"
+#include "llvm/Linker/Linker.h"
+#include "llvm/ProfileData/InstrProf.h"
 #include "llvm/Support/FileSystem.h"
 #include "llvm/Support/Program.h"
 #include "llvm/Support/Path.h"
+#include "llvm/Support/SourceMgr.h"
 #if _WIN32
 #include "llvm/Support/SystemUtils.h"
+#include "llvm/Support/ConvertUTF.h"
 #include <Windows.h>
 #endif
 
@@ -70,18 +75,32 @@ static std::string getOutputName(bool const sharedLib) {
   } else {
     result = "a.out";
   }
+
+  const char *extension = nullptr;
   if (sharedLib) {
-    std::string libExt = std::string(".") + global.dll_ext;
-    if (!endsWith(result, libExt)) {
-      if (global.params.targetTriple.getOS() != llvm::Triple::Win32) {
-        result = "lib" + result + libExt;
-      } else {
-        result.append(libExt);
-      }
+    extension = global.dll_ext;
+    if (global.params.targetTriple->getOS() != llvm::Triple::Win32) {
+      result = "lib" + result;
+    }
+  } else if (global.params.targetTriple->isOSWindows()) {
+    extension = "exe";
+  }
+
+  if (global.params.run) {
+    // If `-run` is passed, the executable is temporary and is removed
+    // after execution. Make sure the name does not collide with other files
+    // from other processes by creating a unique filename.
+    llvm::SmallString<128> tempFilename;
+    auto EC = llvm::sys::fs::createTemporaryFile(
+        result, extension ? extension : "", tempFilename);
+    if (!EC) {
+      result = tempFilename.str();
+    }
+  } else {
+    if (extension) {
+      result += ".";
+      result += extension;
     }
-  } else if (global.params.targetTriple.isOSWindows() &&
-             !endsWith(result, ".exe")) {
-    result.append(".exe");
   }
 
   return result;
@@ -89,6 +108,48 @@ static std::string getOutputName(bool const sharedLib) {
 
 //////////////////////////////////////////////////////////////////////////////
 
+namespace {
+
+#if LDC_LLVM_VER >= 306
+/// Insert an LLVM bitcode file into the module
+void insertBitcodeIntoModule(const char *bcFile, llvm::Module &M,
+                             llvm::LLVMContext &Context) {
+  Logger::println("*** Linking-in bitcode file %s ***", bcFile);
+
+  llvm::SMDiagnostic Err;
+  std::unique_ptr<llvm::Module> loadedModule(
+      getLazyIRFileModule(bcFile, Err, Context));
+  if (!loadedModule) {
+    error(Loc(), "Error when loading LLVM bitcode file: %s", bcFile);
+    fatal();
+  }
+#if LDC_LLVM_VER >= 308
+  llvm::Linker(M).linkInModule(std::move(loadedModule));
+#else
+  llvm::Linker(&M).linkInModule(loadedModule.release());
+#endif
+}
+#endif
+}
+
+/// Insert LLVM bitcode files into the module
+void insertBitcodeFiles(llvm::Module &M, llvm::LLVMContext &Ctx,
+                        Array<const char *> &bitcodeFiles) {
+#if LDC_LLVM_VER >= 306
+  for (const char *fname : bitcodeFiles) {
+    insertBitcodeIntoModule(fname, M, Ctx);
+  }
+#else
+  if (!bitcodeFiles.empty()) {
+    error(Loc(),
+          "Passing LLVM bitcode files to LDC is not supported for LLVM < 3.6");
+    fatal();
+  }
+#endif
+}
+
+//////////////////////////////////////////////////////////////////////////////
+
 static std::string gExePath;
 
 static int linkObjToBinaryGcc(bool sharedLib, bool fullyStatic) {
@@ -106,6 +167,21 @@ static int linkObjToBinaryGcc(bool sharedLib, bool fullyStatic) {
     args.push_back(p);
   }
 
+  // Link with profile-rt library when generating an instrumented binary.
+  // profile-rt uses Phobos (MD5 hashing) and therefore must be passed on the
+  // commandline before Phobos.
+  if (global.params.genInstrProf) {
+#if LDC_LLVM_VER >= 308
+    if (global.params.targetTriple->isOSLinux()) {
+      // For Linux, explicitly define __llvm_profile_runtime as undefined
+      // symbol, so that the initialization part of profile-rt is linked in.
+      args.push_back(
+          ("-Wl,-u," + llvm::getInstrProfRuntimeHookVarName()).str());
+    }
+#endif
+    args.push_back("-lldc-profile-rt");
+  }
+
   // user libs
   for (unsigned i = 0; i < global.params.libfiles->dim; i++) {
     const char *p = static_cast<const char *>(global.params.libfiles->data[i]);
@@ -153,7 +229,11 @@ static int linkObjToBinaryGcc(bool sharedLib, bool fullyStatic) {
     // Don't push -l and -L switches using -Xlinker, but pass them indirectly
     // via GCC. This makes sure user-defined paths take precedence over
     // GCC's builtin LIBRARY_PATHs.
-    if (!p[0] || !(p[0] == '-' && (p[1] == 'l' || p[1] == 'L'))) {
+    // Options starting with -shared and -static are not handled by
+    // the linker and must be passed to the driver.
+    auto str = llvm::StringRef(p);
+    if (!(str.startswith("-l") || str.startswith("-L") ||
+          str.startswith("-shared") || str.startswith("-static"))) {
       args.push_back("-Xlinker");
     }
     args.push_back(p);
@@ -161,13 +241,22 @@ static int linkObjToBinaryGcc(bool sharedLib, bool fullyStatic) {
 
   // default libs
   bool addSoname = false;
-  switch (global.params.targetTriple.getOS()) {
+  switch (global.params.targetTriple->getOS()) {
   case llvm::Triple::Linux:
     addSoname = true;
-    args.push_back("-lrt");
-    if (!opts::disableLinkerStripDead) {
+    // Make sure we don't do --gc-sections when generating a profile-
+    // instrumented binary. The runtime relies on magic sections, which
+    // would be stripped by gc-section on older version of ld, see bug:
+    // https://sourceware.org/bugzilla/show_bug.cgi?id=19161
+    if (!opts::disableLinkerStripDead && !global.params.genInstrProf) {
       args.push_back("-Wl,--gc-sections");
     }
+    if (global.params.targetTriple->getEnvironment() == llvm::Triple::Android) {
+      args.push_back("-ldl");
+      args.push_back("-lm");
+      break;
+    }
+    args.push_back("-lrt");
   // fallthrough
   case llvm::Triple::Darwin:
   case llvm::Triple::MacOSX:
@@ -195,7 +284,7 @@ static int linkObjToBinaryGcc(bool sharedLib, bool fullyStatic) {
     break;
   }
 
-  if (global.params.targetTriple.isWindowsGNUEnvironment()) {
+  if (global.params.targetTriple->isWindowsGNUEnvironment()) {
     // This is really more of a kludge, as linking in the Winsock functions
     // should be handled by the pragma(lib, ...) in std.socket, but it
     // makes LDC behave as expected for now.
@@ -205,13 +294,13 @@ static int linkObjToBinaryGcc(bool sharedLib, bool fullyStatic) {
   // Only specify -m32/-m64 for architectures where the two variants actually
   // exist (as e.g. the GCC ARM toolchain doesn't recognize the switches).
   // MIPS does not have -m32/-m64 but requires -mabi=.
-  if (global.params.targetTriple.get64BitArchVariant().getArch() !=
+  if (global.params.targetTriple->get64BitArchVariant().getArch() !=
           llvm::Triple::UnknownArch &&
-      global.params.targetTriple.get32BitArchVariant().getArch() !=
+      global.params.targetTriple->get32BitArchVariant().getArch() !=
           llvm::Triple::UnknownArch) {
-    if (global.params.targetTriple.get64BitArchVariant().getArch() ==
+    if (global.params.targetTriple->get64BitArchVariant().getArch() ==
             llvm::Triple::mips64 ||
-        global.params.targetTriple.get64BitArchVariant().getArch() ==
+        global.params.targetTriple->get64BitArchVariant().getArch() ==
             llvm::Triple::mips64el) {
       switch (getMipsABI()) {
       case MipsABI::EABI:
@@ -230,7 +319,7 @@ static int linkObjToBinaryGcc(bool sharedLib, bool fullyStatic) {
         break;
       }
     } else {
-      switch (global.params.targetTriple.getArch()) {
+      switch (global.params.targetTriple->getArch()) {
       case llvm::Triple::arm:
       case llvm::Triple::armeb:
       case llvm::Triple::aarch64:
@@ -243,8 +332,7 @@ static int linkObjToBinaryGcc(bool sharedLib, bool fullyStatic) {
       default:
         if (global.params.is64bit) {
           args.push_back("-m64");
-        }
-        else {
+        } else {
           args.push_back("-m32");
         }
       }
@@ -339,9 +427,17 @@ int executeAndWait(const char *commandLine) {
 
   DWORD exitCode;
 
+#if UNICODE
+  std::wstring wcommandLine;
+  if (!llvm::ConvertUTF8toWide(commandLine, wcommandLine))
+    return -3;
+  auto cmdline = const_cast<wchar_t *>(wcommandLine.data());
+#else
+  auto cmdline = const_cast<char *>(commandLine);
+#endif
   // according to MSDN, only CreateProcessW (unicode) may modify the passed
   // command line
-  if (!CreateProcess(NULL, const_cast<char *>(commandLine), NULL, NULL, TRUE, 0,
+  if (!CreateProcess(NULL, cmdline, NULL, NULL, TRUE, 0,
                      NULL, NULL, &si, &pi)) {
     exitCode = -1;
   } else {
@@ -376,12 +472,13 @@ int executeMsvcToolAndWait(const std::string &tool,
 
     auto comspecEnv = getenv("ComSpec");
     if (!comspecEnv) {
-      warning(Loc(), "'ComSpec' environment variable is not set, assuming 'cmd.exe'.");
+      warning(Loc(),
+              "'ComSpec' environment variable is not set, assuming 'cmd.exe'.");
       comspecEnv = "cmd.exe";
     }
     std::string cmdExecutable = comspecEnv;
     std::string batchFile = exe_path::prependBinDir(
-        global.params.targetTriple.isArch64Bit() ? "amd64.bat" : "x86.bat");
+        global.params.targetTriple->isArch64Bit() ? "amd64.bat" : "x86.bat");
 
     commandLine.append(windows::quoteArg(cmdExecutable));
     commandLine.append(" /s /c \"");
@@ -519,6 +616,14 @@ static int linkObjToBinaryWin(bool sharedLib) {
     args.push_back(p);
   }
 
+  // Link with profile-rt library when generating an instrumented binary
+  // profile-rt depends on Phobos (MD5 hashing).
+  if (global.params.genInstrProf) {
+    args.push_back("ldc-profile-rt.lib");
+    // profile-rt depends on ws2_32 for symbol `gethostname`
+    args.push_back("ws2_32.lib");
+  }
+
   // user libs
   for (unsigned i = 0; i < global.params.libfiles->dim; i++) {
     const char *p = static_cast<const char *>(global.params.libfiles->data[i]);
@@ -576,7 +681,7 @@ static int linkObjToBinaryWin(bool sharedLib) {
 //////////////////////////////////////////////////////////////////////////////
 
 int linkObjToBinary(bool sharedLib, bool fullyStatic) {
-  if (global.params.targetTriple.isWindowsMSVCEnvironment()) {
+  if (global.params.targetTriple->isWindowsMSVCEnvironment()) {
     // TODO: Choose dynamic/static MSVCRT version based on fullyStatic?
     return linkObjToBinaryWin(sharedLib);
   }
@@ -590,7 +695,7 @@ int createStaticLibrary() {
   Logger::println("*** Creating static library ***");
 
   const bool isTargetWindows =
-      global.params.targetTriple.isWindowsMSVCEnvironment();
+      global.params.targetTriple->isWindowsMSVCEnvironment();
 
   // find archiver
   std::string tool(isTargetWindows ? "lib.exe" : getArchiver());
