@@ -15,17 +15,30 @@
 #ifndef LDC_DRIVER_LINKER_H
 #define LDC_DRIVER_LINKER_H
 
+namespace llvm {
+class Module;
+class LLVMContext;
+}
+
+template <typename TYPE> struct Array;
+
+/**
+ * Inserts bitcode files passed on the commandline into a module.
+ */
+void insertBitcodeFiles(llvm::Module &M, llvm::LLVMContext &Ctx,
+                        Array<const char *> &bitcodeFiles);
+
 /**
  * Link an executable only from object files.
- * @param argv0 the argv[0] value as passed to main
  * @return 0 on success.
  */
-int linkObjToBinary(bool sharedLib);
+int linkObjToBinary(bool sharedLib, bool fullyStatic);
 
 /**
  * Create a static library from object files.
-*/
-void createStaticLibrary();
+ * @return 0 on success.
+ */
+int createStaticLibrary();
 
 /**
  * Delete the executable that was previously linked with linkObjToBinary.
