@@ -9,17 +9,16 @@
 
 #include "module.h"
 #include "errors.h"
-#include "doc.h"
 #include "id.h"
 #include "hdrgen.h"
 #include "json.h"
 #include "mars.h"
 #include "mtype.h"
-#include "parse.h"
+#include "identifier.h"
 #include "rmem.h"
 #include "root.h"
 #include "scope.h"
-#include "dmd2/target.h"
+#include "ddmd/target.h"
 #include "driver/cl_options.h"
 #include "driver/codegenerator.h"
 #include "driver/configfile.h"
@@ -31,8 +30,10 @@
 #include "gen/irstate.h"
 #include "gen/linkage.h"
 #include "gen/llvm.h"
+#include "gen/llvmhelpers.h"
 #include "gen/logger.h"
 #include "gen/metadata.h"
+#include "gen/objcgen.h"
 #include "gen/optimizer.h"
 #include "gen/passes/Passes.h"
 #include "gen/runtime.h"
@@ -64,8 +65,16 @@
 // Needs Type already declared.
 #include "cond.h"
 
-// in traits.c
-void initTraitsStringTable();
+// From druntime/src/core/runtime.d.
+extern "C" {
+int rt_init();
+}
+
+// In ddmd/doc.d
+void gendocfile(Module *m);
+
+// In driver/main.d
+void writeModuleDependencyFile();
 
 using namespace opts;
 
@@ -103,10 +112,25 @@ static cl::opt<bool> staticFlag(
         "Create a statically linked binary, including all system dependencies"),
     cl::ZeroOrMore);
 
+#if LDC_LLVM_VER >= 309
+static inline llvm::Optional<llvm::Reloc::Model> getRelocModel() {
+  if (mRelocModel.getNumOccurrences()) {
+    llvm::Reloc::Model R = mRelocModel;
+    return R;
+  }
+  return llvm::None;
+}
+#else
+static inline llvm::Reloc::Model getRelocModel() {
+  return mRelocModel;
+}
+#endif
+
 void printVersion() {
   printf("LDC - the LLVM D compiler (%s):\n", global.ldc_version);
   printf("  based on DMD %s and LLVM %s\n", global.version,
          global.llvm_version);
+  printf("  built with %s\n", ldc::built_with_Dcompiler_version);
 #if defined(__has_feature)
 #if __has_feature(address_sanitizer)
   printf("  compiled with address sanitizer enabled\n");
@@ -146,7 +170,7 @@ static void processVersions(std::vector<std::string> &list, const char *type,
       }
     } else {
       char *cstr = mem.xstrdup(value);
-      if (Lexer::isValidIdentifier(cstr)) {
+      if (Identifier::isValidIdentifier(cstr)) {
         addIdent(cstr);
         continue;
       } else {
@@ -156,6 +180,40 @@ static void processVersions(std::vector<std::string> &list, const char *type,
   }
 }
 
+// Helper function to handle -transition=*
+static void processTransitions(std::vector<std::string> &list) {
+    for (const auto &i : list) {
+        if (i == "?") {
+            printf("Language changes listed by -transition=id:\n");
+            printf("  = all           list information on all language changes\n");
+            printf("  = checkimports  give deprecation messages about 10378 anomalies\n");
+            printf("  = complex,14488 list all usages of complex or imaginary types\n");
+            printf("  = field,3449    list all non - mutable fields which occupy an object instance\n");
+            printf("  = import,10378  revert to single phase name lookup\n");
+            printf("  = tls           list all variables going into thread local storage\n");
+            exit(EXIT_SUCCESS);
+        } else if (i == "all") {
+            global.params.vtls = true;
+            global.params.vfield = true;
+            global.params.vcomplex = true;
+            global.params.bug10378 = true; // not set in DMD
+            global.params.check10378 = true; // not set in DMD
+        } else if (i == "checkimports") {
+            global.params.check10378 = true;
+        } else if (i == "complex" || i == "14488") {
+            global.params.vcomplex = true;
+        } else if (i == "field" || i == "3449") {
+            global.params.vfield = true;
+        } else if (i == "import" || i == "10378") {
+            global.params.bug10378 = true;
+        } else if (i == "tls") {
+            global.params.vtls = true;
+        } else {
+            error(Loc(), "Invalid transition %s", i.c_str());
+        }
+    }
+}
+
 // Helper function to handle -of, -od, etc.
 static void initFromString(const char *&dest, const cl::opt<std::string> &src) {
   dest = nullptr;
@@ -274,6 +332,7 @@ static void hideLLVMOptions() {
 #endif
 }
 
+// In driver/main.d
 int main(int argc, char **argv);
 
 static const char *tryGetExplicitConfFile(int argc, char **argv) {
@@ -286,6 +345,34 @@ static const char *tryGetExplicitConfFile(int argc, char **argv) {
   return nullptr;
 }
 
+static llvm::Triple tryGetExplicitTriple(int argc, char **argv) {
+  // most combinations of flags are illegal, this mimicks command line
+  //  behaviour for legal ones only
+  llvm::Triple triple(llvm::sys::getDefaultTargetTriple());
+  const char* mtriple = nullptr;
+  const char* march = nullptr;
+  for (int i = 1; i < argc; ++i) {
+    if (sizeof(void *) != 4 && strcmp(argv[i], "-m32") == 0) {
+      triple = triple.get32BitArchVariant();
+      if (triple.getArch() == llvm::Triple::ArchType::x86)
+        triple.setArchName("i686"); // instead of i386
+      return triple;
+    } else if (sizeof(void *) != 8 && strcmp(argv[i], "-m64") == 0)
+      return triple.get64BitArchVariant();
+    else if (strncmp(argv[i], "-mtriple=", 9) == 0)
+      mtriple = argv[i] + 9;
+    else if (strncmp(argv[i], "-march=", 7) == 0)
+      march = argv[i] + 7;
+  }
+  if (mtriple)
+      triple = llvm::Triple(llvm::Triple::normalize(mtriple));
+  if (march) {
+      std::string errorMsg; // ignore error, will show up later anyway
+      lookupTarget(march, triple, errorMsg); // modifies triple
+  }
+  return triple;
+}
+
 /// Parses switches from the command line, any response files and the global
 /// config file and sets up global.params accordingly.
 ///
@@ -302,6 +389,7 @@ static void parseCommandLine(int argc, char **argv, Strings &sourceFiles,
   global.params.libfiles = new Strings();
   global.params.objfiles = new Strings();
   global.params.ddocfiles = new Strings();
+  global.params.bitcodeFiles = new Strings();
 
   global.params.moduleDeps = nullptr;
   global.params.moduleDepsFile = nullptr;
@@ -312,8 +400,9 @@ static void parseCommandLine(int argc, char **argv, Strings &sourceFiles,
 
   ConfigFile cfg_file;
   const char *explicitConfFile = tryGetExplicitConfFile(argc, argv);
+  std::string cfg_triple = tryGetExplicitTriple(argc, argv).getTriple();
   // just ignore errors for now, they are still printed
-  cfg_file.read(explicitConfFile);
+  cfg_file.read(explicitConfFile, cfg_triple.c_str());
   final_args.insert(final_args.end(), cfg_file.switches_begin(),
                     cfg_file.switches_end());
 
@@ -370,11 +459,33 @@ static void parseCommandLine(int argc, char **argv, Strings &sourceFiles,
     global.params.moduleDeps = new OutBuffer;
   }
 
+// PGO options
+#if LDC_WITH_PGO
+  if (genfileInstrProf.getNumOccurrences() > 0) {
+    global.params.genInstrProf = true;
+    if (genfileInstrProf.empty()) {
+      global.params.datafileInstrProf = "default.profraw";
+    } else {
+      initFromString(global.params.datafileInstrProf, genfileInstrProf);
+    }
+  } else {
+    global.params.genInstrProf = false;
+    // If we don't have to generate instrumentation, we could be given a
+    // profdata file:
+    initFromString(global.params.datafileInstrProf, usefileInstrProf);
+  }
+#else
+  global.params.datafileInstrProf = nullptr;
+  global.params.genInstrProf = false;
+#endif
+
   processVersions(debugArgs, "debug", DebugCondition::setGlobalLevel,
                   DebugCondition::addGlobalIdent);
   processVersions(versions, "version", VersionCondition::setGlobalLevel,
                   VersionCondition::addGlobalIdent);
 
+  processTransitions(transitions);
+
   global.params.output_o =
       (opts::output_o == cl::BOU_UNSET &&
        !(opts::output_bc || opts::output_ll || opts::output_s))
@@ -504,7 +615,11 @@ static void parseCommandLine(int argc, char **argv, Strings &sourceFiles,
     error(Loc(), "-lib and -shared switches cannot be used together");
   }
 
+#if LDC_LLVM_VER >= 309
+  if (createSharedLib && !mRelocModel.getNumOccurrences()) {
+#else
   if (createSharedLib && mRelocModel == llvm::Reloc::Default) {
+#endif
     mRelocModel = llvm::Reloc::PIC_;
   }
 
@@ -608,7 +723,7 @@ static void registerPredefinedFloatABI(const char *soft, const char *hard,
 /// Registers the predefined versions specific to the current target triple
 /// and other target specific options with VersionCondition.
 static void registerPredefinedTargetVersions() {
-  switch (global.params.targetTriple.getArch()) {
+  switch (global.params.targetTriple->getArch()) {
   case llvm::Triple::x86:
     VersionCondition::addPredefinedGlobalIdent("X86");
     if (global.params.useInlineAsm) {
@@ -631,9 +746,9 @@ static void registerPredefinedTargetVersions() {
   case llvm::Triple::ppc64le:
     VersionCondition::addPredefinedGlobalIdent("PPC64");
     registerPredefinedFloatABI("PPC_SoftFloat", "PPC_HardFloat");
-    if (global.params.targetTriple.getOS() == llvm::Triple::Linux) {
+    if (global.params.targetTriple->getOS() == llvm::Triple::Linux) {
       VersionCondition::addPredefinedGlobalIdent(
-          global.params.targetTriple.getArch() == llvm::Triple::ppc64
+          global.params.targetTriple->getArch() == llvm::Triple::ppc64
               ? "ELFv1"
               : "ELFv2");
     }
@@ -696,7 +811,7 @@ static void registerPredefinedTargetVersions() {
     break;
   default:
     error(Loc(), "invalid cpu architecture specified: %s",
-          global.params.targetTriple.getArchName().str().c_str());
+          global.params.targetTriple->getArchName().str().c_str());
     fatal();
   }
 
@@ -719,32 +834,32 @@ static void registerPredefinedTargetVersions() {
   // parse the OS out of the target triple
   // see http://gcc.gnu.org/install/specific.html for details
   // also llvm's different SubTargets have useful information
-  switch (global.params.targetTriple.getOS()) {
+  switch (global.params.targetTriple->getOS()) {
   case llvm::Triple::Win32:
     VersionCondition::addPredefinedGlobalIdent("Windows");
     VersionCondition::addPredefinedGlobalIdent(global.params.is64bit ? "Win64"
                                                                      : "Win32");
-    if (global.params.targetTriple.isKnownWindowsMSVCEnvironment()) {
+    if (global.params.targetTriple->isKnownWindowsMSVCEnvironment()) {
       VersionCondition::addPredefinedGlobalIdent("CRuntime_Microsoft");
     }
-    if (global.params.targetTriple.isWindowsGNUEnvironment()) {
+    if (global.params.targetTriple->isWindowsGNUEnvironment()) {
       VersionCondition::addPredefinedGlobalIdent(
           "mingw32"); // For backwards compatibility.
       VersionCondition::addPredefinedGlobalIdent("MinGW");
     }
-    if (global.params.targetTriple.isWindowsCygwinEnvironment()) {
+    if (global.params.targetTriple->isWindowsCygwinEnvironment()) {
       error(Loc(), "Cygwin is not yet supported");
       fatal();
       VersionCondition::addPredefinedGlobalIdent("Cygwin");
     }
     break;
   case llvm::Triple::Linux:
-    if (global.params.targetTriple.getEnvironment() == llvm::Triple::Android) {
+    VersionCondition::addPredefinedGlobalIdent("linux");
+    VersionCondition::addPredefinedGlobalIdent("Posix");
+    if (global.params.targetTriple->getEnvironment() == llvm::Triple::Android) {
       VersionCondition::addPredefinedGlobalIdent("Android");
       VersionCondition::addPredefinedGlobalIdent("CRuntime_Bionic");
     } else {
-      VersionCondition::addPredefinedGlobalIdent("linux");
-      VersionCondition::addPredefinedGlobalIdent("Posix");
       VersionCondition::addPredefinedGlobalIdent("CRuntime_Glibc");
     }
     break;
@@ -784,13 +899,13 @@ static void registerPredefinedTargetVersions() {
     VersionCondition::addPredefinedGlobalIdent("Posix");
     break;
   default:
-    switch (global.params.targetTriple.getEnvironment()) {
+    switch (global.params.targetTriple->getEnvironment()) {
     case llvm::Triple::Android:
       VersionCondition::addPredefinedGlobalIdent("Android");
       break;
     default:
       error(Loc(), "target '%s' is not yet supported",
-            global.params.targetTriple.str().c_str());
+            global.params.targetTriple->str().c_str());
       fatal();
     }
   }
@@ -807,6 +922,10 @@ static void registerPredefinedVersions() {
     VersionCondition::addPredefinedGlobalIdent("D_Ddoc");
   }
 
+  if (global.params.cov) {
+    VersionCondition::addPredefinedGlobalIdent("D_Coverage");
+  }
+
   if (global.params.useUnitTests) {
     VersionCondition::addPredefinedGlobalIdent("unittest");
   }
@@ -821,6 +940,10 @@ static void registerPredefinedVersions() {
 
   registerPredefinedTargetVersions();
 
+  if (global.params.hasObjectiveC) {
+    VersionCondition::addPredefinedGlobalIdent("D_ObjectiveC");
+  }
+
   // Pass sanitizer arguments to linker. Requires clang.
   if (opts::sanitize == opts::AddressSanitizer) {
     VersionCondition::addPredefinedGlobalIdent("LDC_AddressSanitizer");
@@ -895,7 +1018,7 @@ static void emitJson(Modules &modules) {
 
     ensurePathToNameExists(Loc(), jsonfilename);
 
-    auto jsonfile = new File(jsonfilename);
+    auto jsonfile = File::create(jsonfilename);
 
     jsonfile->setbuffer(buf.data, buf.offset);
     jsonfile->ref = 1;
@@ -903,13 +1026,16 @@ static void emitJson(Modules &modules) {
   }
 }
 
-int main(int argc, char **argv) {
-  // stack trace on signals
+int cppmain(int argc, char **argv) {
+#if LDC_LLVM_VER >= 309
+  llvm::sys::PrintStackTraceOnErrorSignal(argv[0]);
+#else
   llvm::sys::PrintStackTraceOnErrorSignal();
+#endif
 
   exe_path::initialize(argv[0], reinterpret_cast<void *>(main));
 
-  global.init();
+  global._init();
   global.version = ldc::dmd_version;
   global.ldc_version = ldc::ldc_version;
   global.llvm_version = ldc::llvm_version;
@@ -959,7 +1085,7 @@ int main(int argc, char **argv) {
   }
 
   gTargetMachine = createTargetMachine(
-      mTargetTriple, mArch, mCPU, mAttrs, bitness, mFloatABI, mRelocModel,
+      mTargetTriple, mArch, mCPU, mAttrs, bitness, mFloatABI, getRelocModel(),
       mCodeModel, codeGenOptLevel(), disableFpElim, disableLinkerStripDead);
 
 #if LDC_LLVM_VER >= 308
@@ -974,11 +1100,15 @@ int main(int argc, char **argv) {
 #endif
 
   {
-    llvm::Triple triple = llvm::Triple(gTargetMachine->getTargetTriple());
+    llvm::Triple *triple = new llvm::Triple(gTargetMachine->getTargetTriple());
     global.params.targetTriple = triple;
-    global.params.isWindows = triple.isOSWindows();
+    global.params.isWindows = triple->isOSWindows();
     global.params.isLP64 = gDataLayout->getPointerSizeInBits() == 64;
-    global.params.is64bit = triple.isArch64Bit();
+    global.params.is64bit = triple->isArch64Bit();
+    global.params.hasObjectiveC = objc_isSupported(*triple);
+    // mscoff enables slightly different handling of interface functions 
+    // in the front end
+    global.params.mscoff = triple->isKnownWindowsMSVCEnvironment();
   }
 
   // allocate the target abi
@@ -988,7 +1118,7 @@ int main(int argc, char **argv) {
   registerPredefinedVersions();
   dumpPredefinedVersions();
 
-  if (global.params.targetTriple.isOSWindows()) {
+  if (global.params.targetTriple->isOSWindows()) {
     global.dll_ext = "dll";
     global.lib_ext = "lib";
   } else {
@@ -997,15 +1127,13 @@ int main(int argc, char **argv) {
   }
 
   // Initialization
-  Lexer::initLexer();
-  Type::init();
+  Type::_init();
   Id::initialize();
-  Module::init();
-  Target::init();
-  Expression::init();
-  initPrecedence();
+  Module::_init();
+  Target::_init();
+  Expression::_init();
   builtin_init();
-  initTraitsStringTable();
+  objc_init();
 
   // Build import search path
   if (global.params.imppath) {
@@ -1058,17 +1186,27 @@ int main(int argc, char **argv) {
     ext = FileName::ext(p);
     if (ext) {
 #if LDC_POSIX
-      if (strcmp(ext, global.obj_ext) == 0 || strcmp(ext, global.bc_ext) == 0)
+      if (strcmp(ext, global.obj_ext) == 0)
 #else
       if (Port::stricmp(ext, global.obj_ext) == 0 ||
-          Port::stricmp(ext, global.obj_ext_alt) == 0 ||
-          Port::stricmp(ext, global.bc_ext) == 0)
+          Port::stricmp(ext, global.obj_ext_alt) == 0)
 #endif
       {
         global.params.objfiles->push(static_cast<const char *>(files.data[i]));
         continue;
       }
 
+      // Detect LLVM bitcode files on commandline
+#if LDC_POSIX
+      if (strcmp(ext, global.bc_ext) == 0)
+#else
+      if (Port::stricmp(ext, global.bc_ext) == 0)
+#endif
+      {
+        global.params.bitcodeFiles->push(static_cast<const char *>(files.data[i]));
+        continue;
+      }
+
 #if LDC_POSIX
       if (strcmp(ext, "a") == 0)
 #elif __MINGW32__
@@ -1137,9 +1275,9 @@ int main(int argc, char **argv) {
       name = p;
     }
 
-    id = Identifier::idPool(name);
-    auto m = new Module(files.data[i], id, global.params.doDocComments,
-                        global.params.doHdrGeneration);
+    id = Identifier::idPool(name, strlen(name));
+    auto m = Module::create(files.data[i], id, global.params.doDocComments,
+                            global.params.doHdrGeneration);
     modules.push(m);
   }
 
@@ -1163,7 +1301,7 @@ int main(int argc, char **argv) {
     }
 
     m->parse(global.params.doDocComments);
-    m->buildTargetFiles(singleObj, createSharedLib || createStaticLib);
+    buildTargetFiles(m, singleObj, createSharedLib || createStaticLib);
     m->deleteObjFile();
     if (m->isDocFile) {
       gendocfile(m);
@@ -1249,16 +1387,11 @@ int main(int argc, char **argv) {
 
   // Now that we analyzed all modules, write the module dependency file if
   // the user requested it.
-  if (global.params.moduleDepsFile != nullptr) {
-    File deps(global.params.moduleDepsFile);
-    OutBuffer *ob = global.params.moduleDeps;
-    deps.setbuffer(static_cast<void *>(ob->data), ob->offset);
-    deps.write();
-  }
+  writeModuleDependencyFile();
 
   // Generate one or more object/IR/bitcode files.
   if (global.params.obj && !modules.empty()) {
-    ldc::CodeGenerator cg(llvm::getGlobalContext(), singleObj);
+    ldc::CodeGenerator cg(getGlobalContext(), singleObj);
 
     for (unsigned i = 0; i < modules.dim; i++) {
       Module *const m = modules[i];
