@@ -46,7 +46,7 @@ DValue *DtoAAIndex(Loc &loc, Type *type, DValue *aa, DValue *key, bool lvalue) {
   LLFunctionType *funcTy = func->getFunctionType();
 
   // aa param
-  LLValue *aaval = lvalue ? aa->getLVal() : aa->getRVal();
+  LLValue *aaval = lvalue ? DtoLVal(aa) : DtoRVal(aa);
   aaval = DtoBitCast(aaval, funcTy->getParamType(0));
 
   // pkey param
@@ -103,7 +103,7 @@ DValue *DtoAAIndex(Loc &loc, Type *type, DValue *aa, DValue *key, bool lvalue) {
     // if ok, proceed in okbb
     gIR->scope() = IRScope(okbb);
   }
-  return new DVarValue(type, ret);
+  return new DLValue(type, ret);
 }
 
 ////////////////////////////////////////////////////////////////////////////////
@@ -124,7 +124,7 @@ DValue *DtoAAIn(Loc &loc, Type *type, DValue *aa, DValue *key) {
   IF_LOG Logger::cout() << "_aaIn = " << *func << '\n';
 
   // aa param
-  LLValue *aaval = aa->getRVal();
+  LLValue *aaval = DtoRVal(aa);
   IF_LOG {
     Logger::cout() << "aaval: " << *aaval << '\n';
     Logger::cout() << "totype: " << *funcTy->getParamType(0) << '\n';
@@ -170,7 +170,7 @@ DValue *DtoAARemove(Loc &loc, DValue *aa, DValue *key) {
   IF_LOG Logger::cout() << "_aaDel = " << *func << '\n';
 
   // aa param
-  LLValue *aaval = aa->getRVal();
+  LLValue *aaval = DtoRVal(aa);
   IF_LOG {
     Logger::cout() << "aaval: " << *aaval << '\n';
     Logger::cout() << "totype: " << *funcTy->getParamType(0) << '\n';
@@ -194,22 +194,22 @@ DValue *DtoAARemove(Loc &loc, DValue *aa, DValue *key) {
 ////////////////////////////////////////////////////////////////////////////////
 
 LLValue *DtoAAEquals(Loc &loc, TOK op, DValue *l, DValue *r) {
-  Type *t = l->getType()->toBasetype();
-  assert(t == r->getType()->toBasetype() &&
+  Type *t = l->type->toBasetype();
+  assert(t == r->type->toBasetype() &&
          "aa equality is only defined for aas of same type");
   llvm::Function *func =
       getRuntimeFunction(loc, gIR->module, "_aaEqual");
   LLFunctionType *funcTy = func->getFunctionType();
 
-  LLValue *aaval = DtoBitCast(l->getRVal(), funcTy->getParamType(1));
-  LLValue *abval = DtoBitCast(r->getRVal(), funcTy->getParamType(2));
+  LLValue *aaval = DtoBitCast(DtoRVal(l), funcTy->getParamType(1));
+  LLValue *abval = DtoBitCast(DtoRVal(r), funcTy->getParamType(2));
   LLValue *aaTypeInfo = DtoTypeInfoOf(t);
   LLValue *res =
       gIR->CreateCallOrInvoke(func, aaTypeInfo, aaval, abval, "aaEqRes")
           .getInstruction();
-  res = gIR->ir->CreateICmpNE(res, DtoConstInt(0));
-  if (op == TOKnotequal) {
-    res = gIR->ir->CreateNot(res);
-  }
+
+  const auto predicate = eqTokToICmpPred(op, /* invert = */ true);
+  res = gIR->ir->CreateICmp(predicate, res, DtoConstInt(0));
+
   return res;
 }
