@@ -12,6 +12,7 @@
 #include "enum.h"
 #include "id.h"
 #include "init.h"
+#include "nspace.h"
 #include "rmem.h"
 #include "template.h"
 #include "gen/classes.h"
@@ -21,16 +22,12 @@
 #include "gen/llvmhelpers.h"
 #include "gen/logger.h"
 #include "gen/tollvm.h"
+#include "gen/typinf.h"
 #include "gen/uda.h"
 #include "ir/irtype.h"
 #include "ir/irvar.h"
 #include "llvm/ADT/SmallString.h"
 
-//////////////////////////////////////////////////////////////////////////////
-// FIXME: Integrate these functions
-void TypeInfoDeclaration_codegen(TypeInfoDeclaration *decl, IRState *p);
-void TypeInfoClassDeclaration_codegen(TypeInfoDeclaration *decl, IRState *p);
-
 //////////////////////////////////////////////////////////////////////////////
 
 namespace {
@@ -133,24 +130,36 @@ public:
 
   //////////////////////////////////////////////////////////////////////////
 
+  void visit(Nspace *ns) LLVM_OVERRIDE {
+    IF_LOG Logger::println("Nspace::codegen for %s", ns->toPrettyChars());
+    LOG_SCOPE
+
+    if (!isError(ns) && ns->members) {
+      for (auto sym : *ns->members)
+        sym->accept(this);
+    }
+  }
+
+  //////////////////////////////////////////////////////////////////////////
+
   void visit(InterfaceDeclaration *decl) LLVM_OVERRIDE {
     IF_LOG Logger::println("InterfaceDeclaration::codegen: '%s'",
                            decl->toPrettyChars());
     LOG_SCOPE
 
-    if (decl->ir.isDefined()) {
+    if (decl->ir->isDefined()) {
       return;
     }
 
     if (decl->type->ty == Terror) {
       error(decl->loc, "had semantic errors when compiling");
-      decl->ir.setDefined();
+      decl->ir->setDefined();
       return;
     }
 
     if (decl->members && decl->symtab) {
       DtoResolveClass(decl);
-      decl->ir.setDefined();
+      decl->ir->setDefined();
 
       // Emit any members (e.g. final functions).
       for (auto m : *decl->members) {
@@ -175,13 +184,13 @@ public:
                            decl->toPrettyChars());
     LOG_SCOPE
 
-    if (decl->ir.isDefined()) {
+    if (decl->ir->isDefined()) {
       return;
     }
 
     if (decl->type->ty == Terror) {
       error(decl->loc, "had semantic errors when compiling");
-      decl->ir.setDefined();
+      decl->ir->setDefined();
       return;
     }
 
@@ -190,7 +199,7 @@ public:
     }
 
     DtoResolveStruct(decl);
-    decl->ir.setDefined();
+    decl->ir->setDefined();
 
     for (auto m : *decl->members) {
       m->accept(this);
@@ -224,19 +233,19 @@ public:
                            decl->toPrettyChars());
     LOG_SCOPE
 
-    if (decl->ir.isDefined()) {
+    if (decl->ir->isDefined()) {
       return;
     }
 
     if (decl->type->ty == Terror) {
       error(decl->loc, "had semantic errors when compiling");
-      decl->ir.setDefined();
+      decl->ir->setDefined();
       return;
     }
 
     if (decl->members && decl->symtab) {
       DtoResolveClass(decl);
-      decl->ir.setDefined();
+      decl->ir->setDefined();
 
       for (auto m : *decl->members) {
         m->accept(this);
@@ -268,10 +277,10 @@ public:
                            decl->toPrettyChars());
     LOG_SCOPE
 
-    if (decl->ir.isDefined()) {
+    if (decl->ir->isDefined()) {
       return;
     }
-    decl->ir.setDefined();
+    decl->ir->setDefined();
 
     assert(decl->isexp);
     assert(decl->objects);
@@ -290,18 +299,18 @@ public:
                            decl->toPrettyChars());
     LOG_SCOPE;
 
-    if (decl->ir.isDefined()) {
+    if (decl->ir->isDefined()) {
       return;
     }
 
     if (decl->type->ty == Terror) {
       error(decl->loc, "had semantic errors when compiling");
-      decl->ir.setDefined();
+      decl->ir->setDefined();
       return;
     }
 
     DtoResolveVariable(decl);
-    decl->ir.setDefined();
+    decl->ir->setDefined();
 
     // just forward aliases
     if (decl->aliassym) {
@@ -323,11 +332,16 @@ public:
 
       const auto lwc = DtoLinkage(decl);
 
+      if (global.params.vtls && gvar->isThreadLocal() &&
+          !(decl->storage_class & STCtemp)) {
+        const char *p = decl->loc.toChars();
+        fprintf(global.stdmsg, "%s: %s is thread local\n", p, decl->toChars());
+      }
       // Check if we are defining or just declaring the global in this module.
       if (!(decl->storage_class & STCextern)) {
         // Build the initializer. Might use this->ir.irGlobal->value!
         LLConstant *initVal =
-            DtoConstInitializer(decl->loc, decl->type, decl->init);
+            DtoConstInitializer(decl->loc, decl->type, decl->_init);
 
         // In case of type mismatch, swap out the variable.
         if (initVal->getType() != gvar->getType()->getElementType()) {
@@ -400,11 +414,11 @@ public:
                            decl->toPrettyChars());
     LOG_SCOPE
 
-    if (decl->ir.isDefined()) {
+    if (decl->ir->isDefined()) {
       Logger::println("Already defined, skipping.");
       return;
     }
-    decl->ir.setDefined();
+    decl->ir->setDefined();
 
     if (isError(decl)) {
       Logger::println("Has errors, skipping.");
@@ -434,10 +448,10 @@ public:
                            decl->toPrettyChars());
     LOG_SCOPE
 
-    if (decl->ir.isDefined()) {
+    if (decl->ir->isDefined()) {
       return;
     }
-    decl->ir.setDefined();
+    decl->ir->setDefined();
 
     if (!isError(decl) && decl->members) {
       for (auto m : *decl->members) {
@@ -468,18 +482,18 @@ public:
 
       assert(e->op == TOKstring);
       StringExp *se = static_cast<StringExp *>(e);
+      auto name = se->toStringz();
+      auto nameLen = strlen(name);
 
-      size_t nameLen = se->len;
-      if (global.params.targetTriple.isWindowsGNUEnvironment()) {
-        if (nameLen > 4 &&
-            !memcmp(static_cast<char *>(se->string) + nameLen - 4, ".lib", 4)) {
+      if (global.params.targetTriple->isWindowsGNUEnvironment()) {
+        if (nameLen > 4 && !memcmp(name + nameLen - 4, ".lib", 4)) {
           // On MinGW, strip the .lib suffix, if any, to improve
           // compatibility with code written for DMD (we pass the name to GCC
           // via -l, just as on Posix).
           nameLen -= 4;
         }
 
-        if (nameLen >= 7 && !memcmp(se->string, "shell32", 7)) {
+        if (nameLen >= 7 && !memcmp(name, "shell32", 7)) {
           // Another DMD compatibility kludge: Ignore
           // pragma(lib, "shell32.lib"), it is implicitly provided by
           // MinGW.
@@ -489,9 +503,8 @@ public:
 
       // With LLVM 3.3 or later we can place the library name in the object
       // file. This seems to be supported only on Windows.
-      if (global.params.targetTriple.isWindowsMSVCEnvironment()) {
-        llvm::SmallString<24> LibName(
-            llvm::StringRef(static_cast<const char *>(se->string), nameLen));
+      if (global.params.targetTriple->isWindowsMSVCEnvironment()) {
+        llvm::SmallString<24> LibName(llvm::StringRef(name, nameLen));
 
         // Win32: /DEFAULTLIB:"curl"
         if (LibName.endswith(".a")) {
@@ -520,7 +533,7 @@ public:
         char *arg = static_cast<char *>(mem.xmalloc(n));
         arg[0] = '-';
         arg[1] = 'l';
-        memcpy(arg + 2, se->string, nameLen);
+        memcpy(arg + 2, name, nameLen);
         arg[n - 1] = 0;
         global.params.linkswitches->push(arg);
       }
