@@ -21,6 +21,7 @@
 #include "llvm/Support/FileSystem.h"
 #include "llvm/Support/Path.h"
 #include "enum.h"
+#include "ldcbindings.h"
 #include "module.h"
 #include "mtype.h"
 
@@ -162,6 +163,7 @@ ldc::DIType ldc::DIBuilder::CreateEnumType(Type *type) {
   assert(type->ty == Tenum &&
          "only enums allowed for debug info in dwarfEnumType");
   TypeEnum *te = static_cast<TypeEnum *>(type);
+
 #if LDC_LLVM_VER >= 306
   llvm::SmallVector<llvm::Metadata *, 8> subscripts;
 #else
@@ -170,7 +172,7 @@ ldc::DIType ldc::DIBuilder::CreateEnumType(Type *type) {
   for (auto m : *te->sym->members) {
     EnumMember *em = m->isEnumMember();
     llvm::StringRef Name(em->toChars());
-    uint64_t Val = em->value->toInteger();
+    uint64_t Val = em->value()->toInteger();
     auto Subscript = DBuilder.createEnumerator(Name, Val);
     subscripts.push_back(Subscript);
   }
@@ -532,8 +534,20 @@ ldc::DISubroutineType ldc::DIBuilder::CreateDelegateType(Type *type) {
 }
 
 ////////////////////////////////////////////////////////////////////////////////
+bool isOpaqueEnumType(Type *type) {
+  if (type->ty != Tenum)
+    return false;
+
+  TypeEnum *te = static_cast<TypeEnum *>(type);
+  return !te->sym->memtype;
+}
 
 ldc::DIType ldc::DIBuilder::CreateTypeDescription(Type *type, bool derefclass) {
+  // Check for opaque enum first, Bugzilla 13792
+  if (isOpaqueEnumType(type)) {
+    return DBuilder.createUnspecifiedType(type->toChars());
+  }
+
   Type *t = type->toBasetype();
   if (derefclass && t->ty == Tclass) {
     type = type->pointerTo();
@@ -595,7 +609,7 @@ void ldc::DIBuilder::EmitCompileUnit(Module *m) {
   llvm::sys::fs::make_absolute(srcpath);
 
 #if LDC_LLVM_VER >= 308
-  if (global.params.targetTriple.isWindowsMSVCEnvironment())
+  if (global.params.targetTriple->isWindowsMSVCEnvironment())
     IR->module.addModuleFlag(llvm::Module::Warning, "CodeView", 1);
 #endif
   // Metadata without a correct version will be stripped by UpgradeDebugInfo.
@@ -633,21 +647,66 @@ ldc::DISubprogram ldc::DIBuilder::EmitSubProgram(FuncDeclaration *fd) {
 
   // Create subroutine type
   ldc::DISubroutineType DIFnType =
-      CreateFunctionType(static_cast<TypeFunction *>(fd->type));
+    CreateFunctionType(static_cast<TypeFunction *>(fd->type));
 
   // FIXME: duplicates?
   return DBuilder.createFunction(
-      CU,                            // context
-      fd->toPrettyChars(),           // name
-      mangleExact(fd),               // linkage name
-      file,                          // file
-      fd->loc.linnum,                // line no
-      DIFnType,                      // type
-      fd->protection == PROTprivate, // is local to unit
-      true,                          // isdefinition
-      fd->loc.linnum,                // FIXME: scope line
-      DIFlags::FlagPrototyped,       // Flags
-      isOptimizationEnabled()        // isOptimized
+      CU,                                 // context
+      fd->toPrettyChars(),                // name
+      getIrFunc(fd)->func->getName(),     // linkage name
+      file,                               // file
+      fd->loc.linnum,                     // line no
+      DIFnType,                           // type
+      fd->protection.kind == PROTprivate, // is local to unit
+      true,                               // isdefinition
+      fd->loc.linnum,                     // FIXME: scope line
+      DIFlags::FlagPrototyped,            // Flags
+      isOptimizationEnabled()             // isOptimized
+#if LDC_LLVM_VER < 308
+      ,
+      getIrFunc(fd)->func
+#endif
+      );
+}
+
+ldc::DISubprogram ldc::DIBuilder::EmitThunk(llvm::Function *Thunk,
+                                            FuncDeclaration *fd) {
+  if (!global.params.symdebug) {
+#if LDC_LLVM_VER >= 307
+    return nullptr;
+#else
+    return llvm::DISubprogram();
+#endif
+  }
+
+  Logger::println("Thunk to dwarf subprogram");
+  LOG_SCOPE;
+
+  ldc::DICompileUnit CU(GetCU());
+  assert(CU &&
+         "Compilation unit missing or corrupted in DIBuilder::EmitThunk");
+
+  ldc::DIFile file(CreateFile(fd->loc));
+
+  // Create subroutine type (thunk has same type as wrapped function)
+  ldc::DISubroutineType DIFnType = CreateFunctionType(fd->type);
+
+  std::string name = fd->toPrettyChars();
+  name.append(".__thunk");
+
+  // FIXME: duplicates?
+  return DBuilder.createFunction(
+      CU,                                 // context
+      name,                               // name
+      Thunk->getName(),                   // linkage name
+      file,                               // file
+      fd->loc.linnum,                     // line no
+      DIFnType,                           // type
+      fd->protection.kind == PROTprivate, // is local to unit
+      true,                               // isdefinition
+      fd->loc.linnum,                     // FIXME: scope line
+      DIFlags::FlagPrototyped,            // Flags
+      isOptimizationEnabled()             // isOptimized
 #if LDC_LLVM_VER < 308
       ,
       getIrFunc(fd)->func
@@ -969,7 +1028,7 @@ ldc::DIBuilder::EmitGlobalVariable(llvm::GlobalVariable *ll,
   LOG_SCOPE;
 
   assert(vd->isDataseg() ||
-         (vd->storage_class & (STCconst | STCimmutable) && vd->init));
+         (vd->storage_class & (STCconst | STCimmutable) && vd->_init));
 
   return DBuilder.createGlobalVariable(
 #if LDC_LLVM_VER >= 306
@@ -980,7 +1039,7 @@ ldc::DIBuilder::EmitGlobalVariable(llvm::GlobalVariable *ll,
       CreateFile(vd->loc),                    // file
       vd->loc.linnum,                         // line num
       CreateTypeDescription(vd->type, false), // type
-      vd->protection == PROTprivate,          // is local to unit
+      vd->protection.kind == PROTprivate,     // is local to unit
       ll                                      // value
       );
 }
