@@ -26,11 +26,13 @@
 #include <sstream>
 #include <vector>
 #include "llvm/ADT/StringMap.h"
+#include "llvm/ProfileData/InstrProfReader.h"
 #include "llvm/IR/CallSite.h"
 
 namespace llvm {
 class LLVMContext;
 class TargetMachine;
+class IndexedInstrProfReader;
 }
 
 // global ir state for current module
@@ -157,6 +159,10 @@ struct IRState {
   // debug info helper
   ldc::DIBuilder DBuilder;
 
+  // PGO data file reader
+  std::unique_ptr<llvm::IndexedInstrProfReader> PGOReader;
+  llvm::IndexedInstrProfReader *getPGOReader() const { return PGOReader.get(); }
+
   // for inline asm
   IRAsmBlock *asmBlock;
   std::ostringstream nakedAsm;
@@ -182,6 +188,12 @@ struct IRState {
 #else
   llvm::SmallVector<llvm::Value *, 5> LinkerMetadataArgs;
 #endif
+
+#if LDC_LLVM_VER >= 308
+  // MS C++ compatible type descriptors
+  llvm::DenseMap<size_t, llvm::StructType *> TypeDescriptorTypeMap;
+  llvm::DenseMap<llvm::Constant *, llvm::GlobalVariable *> TypeDescriptorMap;
+#endif
 };
 
 void Statement_toIR(Statement *s, IRState *irs);
