@@ -16,11 +16,22 @@
 #ifndef LDC_GEN_LLVMHELPERS_H
 #define LDC_GEN_LLVMHELPERS_H
 
-#include "mtype.h"
-#include "statement.h"
 #include "gen/dvalue.h"
 #include "gen/llvm.h"
 #include "ir/irfuncty.h"
+#include "mtype.h"
+#include "statement.h"
+
+struct IRState;
+
+// An arrayreference type with initializer_list support (C++11):
+#if LDC_LLVM_VER >= 307
+template <class T> using ArrayParam = llvm::ArrayRef<T>;
+#else
+template <class T> using ArrayParam = std::vector<T>;
+#endif
+
+llvm::LLVMContext& getGlobalContext();
 
 // dynamic memory helpers
 LLValue *DtoNew(Loc &loc, Type *newtype);
@@ -118,23 +129,19 @@ LLConstant *DtoConstExpInit(Loc &loc, Type *targetType, Expression *exp);
 LLConstant *DtoTypeInfoOf(Type *ty, bool base = true);
 
 // binary operations
-DValue *DtoBinAdd(DValue *lhs, DValue *rhs);
-DValue *DtoBinSub(DValue *lhs, DValue *rhs);
+DImValue *DtoBinAdd(DRValue *lhs, DRValue *rhs);
+DImValue *DtoBinSub(DRValue *lhs, DRValue *rhs);
 // these binops need an explicit result type to handling
 // to give 'ifloat op float' and 'float op ifloat' the correct type
-DValue *DtoBinMul(Type *resulttype, DValue *lhs, DValue *rhs);
-DValue *DtoBinDiv(Type *resulttype, DValue *lhs, DValue *rhs);
-DValue *DtoBinRem(Type *resulttype, DValue *lhs, DValue *rhs);
+DImValue *DtoBinMul(Type *resulttype, DRValue *lhs, DRValue *rhs);
+DImValue *DtoBinDiv(Type *resulttype, DRValue *lhs, DRValue *rhs);
+DImValue *DtoBinRem(Type *resulttype, DRValue *lhs, DRValue *rhs);
 LLValue *DtoBinNumericEquals(Loc &loc, DValue *lhs, DValue *rhs, TOK op);
 LLValue *DtoBinFloatsEquals(Loc &loc, DValue *lhs, DValue *rhs, TOK op);
 
 // target stuff
 void findDefaultTarget();
 
-/// Fixup an overloaded intrinsic name string.
-void DtoOverloadedIntrinsicName(TemplateInstance *ti, TemplateDeclaration *td,
-                                std::string &name);
-
 /// Returns true if there is any unaligned type inside the aggregate.
 bool hasUnalignedFields(Type *t);
 
@@ -182,9 +189,18 @@ bool isLLVMUnsigned(Type *t);
 ///
 /// For some operations, the result can be a constant. In this case outConst is
 /// set to it, otherwise outPred is set to the predicate to use.
-void tokToIcmpPred(TOK op, bool isUnsigned, llvm::ICmpInst::Predicate *outPred,
+void tokToICmpPred(TOK op, bool isUnsigned, llvm::ICmpInst::Predicate *outPred,
                    llvm::Value **outConst);
 
+/// Converts a DMD equality/identity operation token into the corresponding LLVM
+/// icmp predicate.
+llvm::ICmpInst::Predicate eqTokToICmpPred(TOK op, bool invert = false);
+
+/// For equality/identity operations, returns `(lhs1 == rhs1) & (lhs2 == rhs2)`.
+/// `(lhs1 != rhs1) | (lhs2 != rhs2)` for inequality/not-identity.
+LLValue *createIPairCmp(TOK op, LLValue *lhs1, LLValue *lhs2, LLValue *rhs1,
+                        LLValue *rhs2);
+
 ////////////////////////////////////////////
 // gen/tocall.cpp stuff below
 ////////////////////////////////////////////
@@ -234,6 +250,13 @@ LLConstant *toConstantArray(LLType *ct, LLArrayType *at, T *str, size_t len,
   return LLConstantArray::get(at, vals);
 }
 
+/// Returns the cache for string literals of the given character type (for the
+/// current IRState).
+llvm::StringMap<llvm::GlobalVariable *> *
+stringLiteralCacheForType(Type *charType);
+
+llvm::Constant *buildStringLiteralConstant(StringExp *se, bool zeroTerm);
+
 /// Tries to create an LLVM global with the given properties. If a variable with
 /// the same mangled name already exists, checks if the types match and returns
 /// it instead.
@@ -257,4 +280,14 @@ DValue *toElem(Expression *e, bool tryGetLvalue);
 DValue *toElemDtor(Expression *e);
 LLConstant *toConstElem(Expression *e, IRState *p);
 
+inline llvm::Value *DtoRVal(Expression *e) { return DtoRVal(toElem(e)); }
+inline llvm::Value *DtoLVal(Expression *e) { return DtoLVal(toElem(e)); }
+
+/// Creates a DLValue for the given VarDeclaration.
+///
+/// If the storage is not given explicitly, the declaration is expected to be
+/// already resolved, and the value from the associated IrVar will be used.
+DValue *makeVarDValue(Type *type, VarDeclaration *vd,
+                      llvm::Value *storage = nullptr);
+
 #endif
