@@ -29,6 +29,7 @@
 #include "gen/llvm.h"
 #include "gen/llvmhelpers.h"
 #include "gen/logger.h"
+#include "gen/objcgen.h"
 #include "gen/optimizer.h"
 #include "gen/programs.h"
 #include "gen/rttibuilder.h"
@@ -41,11 +42,15 @@
 #include "ir/irtype.h"
 #include "ir/irvar.h"
 #include "llvm/Support/CommandLine.h"
+#include "llvm/Support/FileSystem.h"
+#include "llvm/Support/Path.h"
+#include "llvm/IR/DataLayout.h"
+#include "llvm/IR/Module.h"
 #include "llvm/IR/Verifier.h"
 #include "llvm/LinkAllPasses.h"
+#include "llvm/ProfileData/InstrProfReader.h"
+#include "llvm/Support/CommandLine.h"
 #include "llvm/Transforms/Utils/ModuleUtils.h"
-#include "llvm/IR/Module.h"
-#include "llvm/IR/DataLayout.h"
 
 #if _AIX || __sun
 #include <alloca.h>
@@ -74,111 +79,97 @@ static void check_and_add_output_file(Module *NewMod, const std::string &str) {
   files.insert(std::make_pair(str, NewMod));
 }
 
-void Module::buildTargetFiles(bool singleObj, bool library) {
-  if (objfile && (!doDocComment || docfile) && (!doHdrGen || hdrfile)) {
+void buildTargetFiles(Module *m, bool singleObj, bool library) {
+  if (m->objfile && (!m->doDocComment || m->docfile) &&
+      (!m->doHdrGen || m->hdrfile)) {
     return;
   }
 
-  if (!objfile) {
+  if (!m->objfile) {
     const char *objname = library ? nullptr : global.params.objname;
     if (global.params.output_o) {
-      objfile = Module::buildFilePath(objname, global.params.objdir,
-                                      global.params.targetTriple.isOSWindows()
-                                          ? global.obj_ext_alt
-                                          : global.obj_ext);
+      const char *extension = global.params.targetTriple->isOSWindows()
+                                  ? global.obj_ext_alt
+                                  : global.obj_ext;
+      llvm::SmallString<128> tempFilename; // must outlive buildFilePath call
+      if (global.params.run) {
+        using namespace llvm;
+        // If `-run` is passed, the obj file is temporary and is removed
+        // after execution. Make sure the name does not collide with other files
+        // from other processes by creating a unique filename.
+        auto tmpname = objname ? objname : "tmp";
+        std::error_code EC;
+        if (global.params.objdir) {
+          // Prepend with path to form absolute filename.
+          EC = sys::fs::createUniqueFile(Twine(global.params.objdir) +
+                                             sys::path::get_separator() +
+                                             tmpname + "-%%%%%%%." + extension,
+                                         tempFilename);
+        } else {
+          // Uses current dir as base for file.
+          EC = sys::fs::createUniqueFile(
+              Twine(tmpname) + "-%%%%%%%." + extension, tempFilename);
+        }
+        if (!EC) {
+          objname = tempFilename.c_str();
+        }
+      }
+      m->objfile = m->buildFilePath(objname, global.params.objdir, extension,
+                                    library, fqnNames);
     } else if (global.params.output_bc) {
-      objfile =
-          Module::buildFilePath(objname, global.params.objdir, global.bc_ext);
+      m->objfile = m->buildFilePath(objname, global.params.objdir,
+                                    global.bc_ext, library, fqnNames);
     } else if (global.params.output_ll) {
-      objfile =
-          Module::buildFilePath(objname, global.params.objdir, global.ll_ext);
+      m->objfile = m->buildFilePath(objname, global.params.objdir,
+                                    global.ll_ext, library, fqnNames);
     } else if (global.params.output_s) {
-      objfile =
-          Module::buildFilePath(objname, global.params.objdir, global.s_ext);
+      m->objfile = m->buildFilePath(objname, global.params.objdir, global.s_ext,
+                                    library, fqnNames);
     }
   }
-  if (doDocComment && !docfile) {
-    docfile = Module::buildFilePath(global.params.docname, global.params.docdir,
-                                    global.doc_ext);
+  if (m->doDocComment && !m->docfile) {
+    m->docfile = m->buildFilePath(global.params.docname, global.params.docdir,
+                                  global.doc_ext, library, fqnNames);
   }
-  if (doHdrGen && !hdrfile) {
-    hdrfile = Module::buildFilePath(global.params.hdrname, global.params.hdrdir,
-                                    global.hdr_ext);
+  if (m->doHdrGen && !m->hdrfile) {
+    m->hdrfile = m->buildFilePath(global.params.hdrname, global.params.hdrdir,
+                                  global.hdr_ext, library, fqnNames);
   }
 
   // safety check: never allow obj, doc or hdr file to have the source file's
   // name
-  if (Port::stricmp(FileName::name(objfile->name->str),
-                    FileName::name(this->arg)) == 0) {
-    error("Output object files with the same name as the source file are "
-          "forbidden");
+  if (Port::stricmp(FileName::name(m->objfile->name->str),
+                    FileName::name(m->arg)) == 0) {
+    m->error("Output object files with the same name as the source file are "
+             "forbidden");
     fatal();
   }
-  if (docfile &&
-      Port::stricmp(FileName::name(docfile->name->str),
-                    FileName::name(this->arg)) == 0) {
-    error(
+  if (m->docfile &&
+      Port::stricmp(FileName::name(m->docfile->name->str),
+                    FileName::name(m->arg)) == 0) {
+    m->error(
         "Output doc files with the same name as the source file are forbidden");
     fatal();
   }
-  if (hdrfile &&
-      Port::stricmp(FileName::name(hdrfile->name->str),
-                    FileName::name(this->arg)) == 0) {
-    error("Output header files with the same name as the source file are "
-          "forbidden");
+  if (m->hdrfile &&
+      Port::stricmp(FileName::name(m->hdrfile->name->str),
+                    FileName::name(m->arg)) == 0) {
+    m->error("Output header files with the same name as the source file are "
+             "forbidden");
     fatal();
   }
 
   // LDC
   // another safety check to make sure we don't overwrite previous output files
   if (!singleObj && global.params.obj) {
-    check_and_add_output_file(this, objfile->name->str);
+    check_and_add_output_file(m, m->objfile->name->str);
   }
-  if (docfile) {
-    check_and_add_output_file(this, docfile->name->str);
+  if (m->docfile) {
+    check_and_add_output_file(m, m->docfile->name->str);
   }
   // FIXME: DMD overwrites header files. This should be done only in a DMD mode.
   // if (hdrfile)
-  //    check_and_add_output_file(this, hdrfile->name->str);
-}
-
-File *Module::buildFilePath(const char *forcename, const char *path,
-                            const char *ext) {
-  const char *argobj;
-  if (forcename) {
-    argobj = forcename;
-  } else {
-    if (preservePaths) {
-      argobj = this->arg;
-    } else {
-      argobj = FileName::name(this->arg);
-    }
-
-    if (fqnNames) {
-      char *name = md ? md->toChars() : toChars();
-      argobj = FileName::replaceName(argobj, name);
-
-      // add ext, otherwise forceExt will make nested.module into nested.bc
-      size_t len = strlen(argobj);
-      size_t extlen = strlen(ext);
-      char *s = reinterpret_cast<char *>(alloca(len + 1 + extlen + 1));
-      memcpy(s, argobj, len);
-      s[len] = '.';
-      memcpy(s + len + 1, ext, extlen + 1);
-      s[len + 1 + extlen] = 0;
-      argobj = s;
-    }
-  }
-
-  if (!FileName::absolute(argobj)) {
-    argobj = FileName::combine(path, argobj);
-  }
-
-  FileName::ensurePathExists(FileName::path(argobj));
-
-  // always append the extension! otherwise hard to make output switches
-  // consistent
-  return new File(FileName::forceExt(argobj, ext));
+  //    check_and_add_output_file(m, hdrfile->name->str);
 }
 
 static llvm::Function *build_module_function(
@@ -198,7 +189,7 @@ static llvm::Function *build_module_function(
   LLFunctionType *fnTy = LLFunctionType::get(LLType::getVoidTy(gIR->context()),
                                              std::vector<LLType *>(), false);
 
-  std::string const symbolName = gABI->mangleForLLVM(name, LINKd);
+  std::string const symbolName = gABI->mangleFunctionForLLVM(name, LINKd);
   assert(gIR->module.getFunction(symbolName) == NULL);
   llvm::Function *fn = llvm::Function::Create(
       fnTy, llvm::GlobalValue::InternalLinkage, symbolName, &gIR->module);
@@ -618,7 +609,8 @@ static void addCoverageAnalysis(Module *m) {
     // Assert that initializer array elements have enough bits
     assert(sizeof(m->d_cover_valid_init[0]) * 8 >=
            gDataLayout->getTypeSizeInBits(DtoSize_t()));
-    m->d_cover_valid_init.resize(array_size);
+    m->d_cover_valid_init.setDim(array_size);
+    m->d_cover_valid_init.zero();
   }
 
   // uint[# source lines] _d_cover_data
@@ -661,7 +653,7 @@ static void addCoverageAnalysis(Module *m) {
                               &gIR->module);
     ctor->setCallingConv(gABI->callingConv(ctor->getFunctionType(), LINKd));
     // Set function attributes. See functions.cpp:DtoDefineFunction()
-    if (global.params.targetTriple.getArch() == llvm::Triple::x86_64) {
+    if (global.params.targetTriple->getArch() == llvm::Triple::x86_64) {
       ctor->addFnAttr(LLAttribute::UWTable);
     }
 
@@ -714,6 +706,60 @@ static void addCoverageAnalysisInitializer(Module *m) {
   m->d_cover_valid->setInitializer(llvm::ConstantArray::get(type, arrayInits));
 }
 
+// Load InstrProf data from file and store in it IrState
+// TODO: This is probably not the right place, we should load it once for all
+// modules?
+static void loadInstrProfileData(IRState *irs) {
+#if LDC_WITH_PGO
+  // Only load from datafileInstrProf if we are not generating instrumented
+  // code.
+  if (!global.params.genInstrProf && global.params.datafileInstrProf) {
+    IF_LOG Logger::println("Read profile data from %s",
+                           global.params.datafileInstrProf);
+
+    auto readerOrErr =
+        llvm::IndexedInstrProfReader::create(global.params.datafileInstrProf);
+#if LDC_LLVM_VER >= 309
+    if (auto E = readerOrErr.takeError()) {
+      handleAllErrors(std::move(E), [&](const llvm::ErrorInfoBase &EI) {
+        irs->dmodule->error("Could not read profile file %s: %s",
+                            global.params.datafileInstrProf,
+                            EI.message().c_str());
+      });
+      fatal();
+    }
+#else
+    std::error_code EC = readerOrErr.getError();
+    if (EC) {
+      irs->dmodule->error("Could not read profile file %s: %s",
+                          global.params.datafileInstrProf,
+                          EC.message().c_str());
+      fatal();
+    }
+#endif
+    irs->PGOReader = std::move(readerOrErr.get());
+
+#if LDC_LLVM_VER >= 309
+    if (!irs->module.getProfileSummary()) {
+      // Don't reset the summary. There is only one profile data file per LDC
+      // invocation so the summary must be the same as the one that is already
+      // set.
+      irs->module.setProfileSummary(
+          irs->PGOReader->getSummary().getMD(irs->context()));
+    }
+#elif LDC_LLVM_VER == 308
+    auto maxCount = irs->PGOReader->getMaximumFunctionCount();
+    if (!irs->module.getMaximumFunctionCount()) {
+      // Don't reset the max function count. There is only one profile data file
+      // per LDC invocation so the information must be the same as the one that
+      // is already set.
+      irs->module.setMaximumFunctionCount(maxCount);
+    }
+#endif
+  }
+#endif
+}
+
 static void genModuleInfo(Module *m, bool emitFullModuleInfo);
 
 void codegenModule(IRState *irs, Module *m, bool emitFullModuleInfo) {
@@ -727,10 +773,15 @@ void codegenModule(IRState *irs, Module *m, bool emitFullModuleInfo) {
 
   // Skip pseudo-modules for coverage analysis
   std::string name = m->toChars();
-  if (global.params.cov && name != "__entrypoint" && name != "__main") {
+  bool pseudo_modules = (name == "__entrypoint") || (name == "__main");
+  if (global.params.cov && !pseudo_modules) {
     addCoverageAnalysis(m);
   }
 
+  if (!pseudo_modules) {
+    loadInstrProfileData(gIR);
+  }
+
   // process module members
   for (unsigned k = 0; k < m->members->dim; k++) {
     Dsymbol *dsym = (*m->members)[k];
@@ -930,21 +981,24 @@ static void genModuleInfo(Module *m, bool emitFullModuleInfo) {
   llvm::ArrayType *at = llvm::ArrayType::get(it, len);
   b.push(toConstantArray(it, at, name, len, false));
 
+  objc_Module_genmoduleinfo_classes();
+
   // create and set initializer
   LLGlobalVariable *moduleInfoSym = getIrModule(m)->moduleInfoSymbol();
   b.finalize(moduleInfoSym->getType()->getPointerElementType(), moduleInfoSym);
   setLinkage({LLGlobalValue::ExternalLinkage, false}, moduleInfoSym);
 
-  if (global.params.targetTriple.isOSLinux() ||
-      global.params.targetTriple.isOSFreeBSD() ||
+  if ((global.params.targetTriple->isOSLinux() &&
+       global.params.targetTriple->getEnvironment() != llvm::Triple::Android) ||
+      global.params.targetTriple->isOSFreeBSD() ||
 #if LDC_LLVM_VER > 305
-      global.params.targetTriple.isOSNetBSD() ||
-      global.params.targetTriple.isOSOpenBSD() ||
-      global.params.targetTriple.isOSDragonFly()
+      global.params.targetTriple->isOSNetBSD() ||
+      global.params.targetTriple->isOSOpenBSD() ||
+      global.params.targetTriple->isOSDragonFly()
 #else
-      global.params.targetTriple.getOS() == llvm::Triple::NetBSD ||
-      global.params.targetTriple.getOS() == llvm::Triple::OpenBSD ||
-      global.params.targetTriple.getOS() == llvm::Triple::DragonFly
+      global.params.targetTriple->getOS() == llvm::Triple::NetBSD ||
+      global.params.targetTriple->getOS() == llvm::Triple::OpenBSD ||
+      global.params.targetTriple->getOS() == llvm::Triple::DragonFly
 #endif
           ) {
     if (emitFullModuleInfo) {
