@@ -19,23 +19,6 @@
 #include "ir/irtypeaggr.h"
 #include "llvm/Analysis/ValueTracking.h"
 
-static void storeVariable(VarDeclaration *vd, LLValue *dst) {
-  LLValue *value = getIrLocal(vd)->value;
-  int ty = vd->type->ty;
-  FuncDeclaration *fd = getParentFunc(vd, true);
-  assert(fd && "No parent function for nested variable?");
-  if (fd->needsClosure() && !vd->isRef() && (ty == Tstruct || ty == Tsarray) &&
-      isaPointer(value->getType())) {
-    // Copy structs and static arrays
-    LLValue *mem = DtoGcMalloc(vd->loc, DtoType(vd->type), ".gc_mem");
-    DtoMemCpy(mem, value);
-    DtoAlignedStore(mem, dst);
-  } else {
-    // Store the address into the frame
-    DtoAlignedStore(value, dst);
-  }
-}
-
 static unsigned getVthisIdx(AggregateDeclaration *ad) {
   return getFieldGEPIndex(ad, ad->vthis);
 }
@@ -62,8 +45,7 @@ DValue *DtoNestedVariable(Loc &loc, Type *astype, VarDeclaration *vd,
     if (fd->isStatic()) {
       error(loc, "function %s cannot access frame of function %s",
             irfunc->decl->toPrettyChars(), vdparent->toPrettyChars());
-      return new DVarValue(astype, vd,
-                           llvm::UndefValue::get(DtoPtrToType(astype)));
+      return new DLValue(astype, llvm::UndefValue::get(DtoPtrToType(astype)));
     }
     fd = getParentFunc(fd, false);
     assert(fd);
@@ -71,8 +53,7 @@ DValue *DtoNestedVariable(Loc &loc, Type *astype, VarDeclaration *vd,
 
   // is the nested variable in this scope?
   if (vdparent == irfunc->decl) {
-    LLValue *val = getIrValue(vd);
-    return new DVarValue(astype, vd, val);
+    return makeVarDValue(astype, vd);
   }
 
   LLValue *dwarfValue = nullptr;
@@ -85,12 +66,14 @@ DValue *DtoNestedVariable(Loc &loc, Type *astype, VarDeclaration *vd,
   // get the nested context
   LLValue *ctx = nullptr;
   if (irfunc->nestedVar) {
-    // If this function has its own nested context struct, always load it.
+    Logger::println("Using own nested context of current function");
+
     ctx = irfunc->nestedVar;
     dwarfValue = ctx;
   } else if (irfunc->decl->isMember2()) {
-    // If this is a member function of a nested class without its own
-    // context, load the vthis member.
+    Logger::println(
+        "Current function is member of nested class, loading vthis");
+
     AggregateDeclaration *cd = irfunc->decl->isMember2();
     LLValue *val = irfunc->thisArg;
     if (cd->isClassDeclaration()) {
@@ -98,15 +81,17 @@ DValue *DtoNestedVariable(Loc &loc, Type *astype, VarDeclaration *vd,
     }
     ctx = DtoLoad(DtoGEPi(val, 0, getVthisIdx(cd), ".vthis"));
   } else {
-    // Otherwise, this is a simple nested function, load from the context
-    // argument.
+    Logger::println("Regular nested function, loading context arg");
+
     ctx = DtoLoad(irfunc->nestArg);
     dwarfValue = irfunc->nestArg;
     if (global.params.symdebug) {
       gIR->DBuilder.OpDeref(dwarfAddr);
     }
   }
+
   assert(ctx);
+  IF_LOG { Logger::cout() << "Context: " << *ctx << '\n'; }
 
   DtoCreateNestedContextType(vdparent->isFuncDeclaration());
   assert(isIrLocalCreated(vd));
@@ -114,14 +99,13 @@ DValue *DtoNestedVariable(Loc &loc, Type *astype, VarDeclaration *vd,
   ////////////////////////////////////
   // Extract variable from nested context
 
-  LLValue *val = DtoBitCast(ctx, LLPointerType::getUnqual(irfunc->frameType));
-  IF_LOG {
-    Logger::cout() << "Context: " << *val << '\n';
-    Logger::cout() << "of type: " << *irfunc->frameType << '\n';
-  }
+  const auto frameType = LLPointerType::getUnqual(irfunc->frameType);
+  IF_LOG { Logger::cout() << "casting to: " << *irfunc->frameType << '\n'; }
+  LLValue *val = DtoBitCast(ctx, frameType);
 
-  unsigned vardepth = getIrLocal(vd)->nestedDepth;
-  unsigned funcdepth = irfunc->depth;
+  IrLocal *const irLocal = getIrLocal(vd);
+  const auto vardepth = irLocal->nestedDepth;
+  const auto funcdepth = irfunc->depth;
 
   IF_LOG {
     Logger::cout() << "Variable: " << vd->toChars() << '\n';
@@ -148,7 +132,7 @@ DValue *DtoNestedVariable(Loc &loc, Type *astype, VarDeclaration *vd,
     IF_LOG Logger::cout() << "Frame: " << *val << '\n';
   }
 
-  int idx = getIrLocal(vd)->nestedIndex;
+  const auto idx = irLocal->nestedIndex;
   assert(idx != -1 && "Nested context not yet resolved for variable.");
 
   if (dwarfValue && global.params.symdebug) {
@@ -160,13 +144,12 @@ DValue *DtoNestedVariable(Loc &loc, Type *astype, VarDeclaration *vd,
     Logger::cout() << "Addr: " << *val << '\n';
     Logger::cout() << "of type: " << *val->getType() << '\n';
   }
-  if (byref || (vd->isParameter() && getIrParameter(vd)->arg &&
-                getIrParameter(vd)->arg->byref)) {
+  if (!isSpecialRefVar(vd) && (byref || vd->isRef() || vd->isOut())) {
     val = DtoAlignedLoad(val);
     // dwarfOpDeref(dwarfAddr);
     IF_LOG {
-      Logger::cout() << "Was byref, now: " << *val << '\n';
-      Logger::cout() << "of type: " << *val->getType() << '\n';
+      Logger::cout() << "Was byref, now: " << *irLocal->value << '\n';
+      Logger::cout() << "of type: " << *irLocal->value->getType() << '\n';
     }
   }
 
@@ -174,7 +157,7 @@ DValue *DtoNestedVariable(Loc &loc, Type *astype, VarDeclaration *vd,
     gIR->DBuilder.EmitLocalVariable(dwarfValue, vd, nullptr, false, dwarfAddr);
   }
 
-  return new DVarValue(astype, vd, val);
+  return makeVarDValue(astype, vd, val);
 }
 
 void DtoResolveNestedContext(Loc &loc, AggregateDeclaration *decl,
@@ -328,6 +311,7 @@ static void DtoCreateNestedContextType(FuncDeclaration *fd) {
   IrFunction &irFunc = *getIrFunc(fd);
 
   if (irFunc.nestedContextCreated) {
+    Logger::println("already done");
     return;
   }
   irFunc.nestedContextCreated = true;
@@ -338,98 +322,8 @@ static void DtoCreateNestedContextType(FuncDeclaration *fd) {
     DtoCreateNestedContextType(parentFunc);
   }
 
-  // construct nested variables array
-  if (fd->closureVars.dim > 0) {
-    Logger::println("has nested frame");
-    // start with adding all enclosing parent frames until a static parent is
-    // reached
-
-    LLStructType *innerFrameType = nullptr;
-    unsigned depth = 0;
-
-    if (parentFunc) {
-      IrFunction &parentIrFunc = *getIrFunc(parentFunc);
-      innerFrameType = parentIrFunc.frameType;
-      if (innerFrameType) {
-        depth = parentIrFunc.depth + 1;
-      }
-    }
-
-    irFunc.depth = depth;
-
-    IF_LOG Logger::cout() << "Function " << fd->toChars() << " has depth "
-                          << depth << '\n';
-
-    AggrTypeBuilder builder(false);
-
-    if (depth != 0) {
-      assert(innerFrameType);
-      unsigned ptrSize = gDataLayout->getPointerSize();
-      // Add frame pointer types for all but last frame
-      for (unsigned i = 0; i < (depth - 1); ++i) {
-        builder.addType(innerFrameType->getElementType(i), ptrSize);
-      }
-      // Add frame pointer type for last frame
-      builder.addType(LLPointerType::getUnqual(innerFrameType), ptrSize);
-    }
-
-    // Add the direct nested variables of this function, and update their
-    // indices to match.
-    // TODO: optimize ordering for minimal space usage?
-    for (auto vd : fd->closureVars) {
-      unsigned alignment = DtoAlignment(vd);
-      if (alignment > 1) {
-        builder.alignCurrentOffset(alignment);
-      }
-
-      IrLocal &irLocal = *getIrLocal(vd, true);
-      irLocal.nestedIndex = builder.currentFieldIndex();
-      irLocal.nestedDepth = depth;
-
-      LLType *t = nullptr;
-      if (vd->isParameter() && getIrParameter(vd)->arg) {
-        // Parameters that are part of the LLVM signature will have
-        // storage associated with them (to handle byref etc.), so
-        // handle those cases specially by storing a pointer instead
-        // of a value.
-        const IrParameter *irparam = getIrParameter(vd);
-        const bool refout = vd->storage_class & (STCref | STCout);
-        const bool lazy = vd->storage_class & STClazy;
-        const bool byref = irparam->arg->byref;
-        const bool isVthisPtr = irparam->isVthis && !byref;
-        if (!(refout || (byref && !lazy)) || isVthisPtr) {
-          // This will be copied to the nesting frame.
-          if (lazy) {
-            t = irparam->value->getType()->getContainedType(0);
-          } else {
-            t = DtoMemType(vd->type);
-          }
-        } else {
-          t = irparam->value->getType();
-        }
-      } else if (isSpecialRefVar(vd)) {
-        t = DtoType(vd->type->pointerTo());
-      } else {
-        t = DtoMemType(vd->type);
-      }
-
-      builder.addType(t, getTypeAllocSize(t));
-
-      IF_LOG Logger::cout() << "Nested var '" << vd->toChars() << "' of type "
-                            << *t << "\n";
-    }
-
-    LLStructType *frameType =
-        LLStructType::create(gIR->context(), builder.defaultTypes(),
-                             std::string("nest.") + fd->toChars());
-
-    IF_LOG Logger::cout() << "frameType = " << *frameType << '\n';
-
-    // Store type in IrFunction
-    irFunc.frameType = frameType;
-    irFunc.frameTypeAlignment = builder.overallAlignment();
-  } else // no captured variables
-  {
+  if (fd->closureVars.dim == 0) {
+    // No local variables of this function are captured.
     if (parentFunc) {
       // Propagate context arg properties if the context arg is passed on
       // unmodified.
@@ -438,7 +332,80 @@ static void DtoCreateNestedContextType(FuncDeclaration *fd) {
       irFunc.frameTypeAlignment = parentIrFunc.frameTypeAlignment;
       irFunc.depth = parentIrFunc.depth;
     }
+    return;
   }
+
+  Logger::println("has nested frame");
+
+  // construct nested variables array
+  // start with adding all enclosing parent frames until a static parent is
+  // reached
+
+  LLStructType *innerFrameType = nullptr;
+  unsigned depth = 0;
+
+  if (parentFunc) {
+    IrFunction &parentIrFunc = *getIrFunc(parentFunc);
+    innerFrameType = parentIrFunc.frameType;
+    if (innerFrameType) {
+      depth = parentIrFunc.depth + 1;
+    }
+  }
+
+  irFunc.depth = depth;
+
+  IF_LOG Logger::cout() << "Function " << fd->toChars() << " has depth "
+                        << depth << '\n';
+
+  AggrTypeBuilder builder(false);
+
+  if (depth != 0) {
+    assert(innerFrameType);
+    unsigned ptrSize = gDataLayout->getPointerSize();
+    // Add frame pointer types for all but last frame
+    for (unsigned i = 0; i < (depth - 1); ++i) {
+      builder.addType(innerFrameType->getElementType(i), ptrSize);
+    }
+    // Add frame pointer type for last frame
+    builder.addType(LLPointerType::getUnqual(innerFrameType), ptrSize);
+  }
+
+  // Add the direct nested variables of this function, and update their
+  // indices to match.
+  // TODO: optimize ordering for minimal space usage?
+  for (auto vd : fd->closureVars) {
+    unsigned alignment = DtoAlignment(vd);
+    if (alignment > 1) {
+      builder.alignCurrentOffset(alignment);
+    }
+
+    IrLocal &irLocal = *getIrLocal(vd, true);
+    irLocal.nestedIndex = builder.currentFieldIndex();
+    irLocal.nestedDepth = depth;
+
+    LLType *t = nullptr;
+    if (vd->isRef() || vd->isOut())
+      t = DtoType(vd->type->pointerTo());
+    else if (vd->isParameter() && (vd->storage_class & STClazy))
+      t = getIrParameter(vd)->value->getType()->getContainedType(0);
+    else
+      t = DtoMemType(vd->type);
+
+    builder.addType(t, getTypeAllocSize(t));
+
+    IF_LOG Logger::cout() << "Nested var '" << vd->toChars() << "' of type "
+                          << *t << "\n";
+  }
+
+  LLStructType *frameType =
+      LLStructType::create(gIR->context(), builder.defaultTypes(),
+                           std::string("nest.") + fd->toChars());
+
+  IF_LOG Logger::cout() << "frameType = " << *frameType << '\n';
+
+  // Store type in IrFunction
+  irFunc.frameType = frameType;
+  irFunc.frameTypeAlignment = builder.overallAlignment();
 }
 
 void DtoCreateNestedContext(FuncDeclaration *fd) {
@@ -501,7 +468,7 @@ void DtoCreateNestedContext(FuncDeclaration *fd) {
 
     // go through all nested vars and assign addresses where possible.
     for (auto vd : fd->closureVars) {
-      if (needsClosure && vd->needsAutoDtor()) {
+      if (needsClosure && vd->needsScopeDtor()) {
         // This should really be a front-end, not a glue layer error,
         // but we need to fix this in DMD too.
         vd->error("has scoped destruction, cannot build closure");
@@ -513,9 +480,12 @@ void DtoCreateNestedContext(FuncDeclaration *fd) {
         IF_LOG Logger::println("nested param: %s", vd->toChars());
         LOG_SCOPE
         IrParameter *parm = getIrParameter(vd);
+        assert(parm->value);
+        assert(parm->value->getType()->isPointerTy());
 
-        if (parm->arg && parm->arg->byref) {
-          storeVariable(vd, gep);
+        if (vd->isRef() || vd->isOut()) {
+          Logger::println("Captured by reference, copying pointer to nested frame");
+          DtoAlignedStore(parm->value, gep);
         } else {
           Logger::println("Copying to nested frame");
           // The parameter value is an alloca'd stack slot.
