@@ -8,7 +8,7 @@
 //===----------------------------------------------------------------------===//
 
 #include "gen/optimizer.h"
-#include "mars.h" // error()
+#include "errors.h"
 #include "gen/cl_helpers.h"
 #include "gen/logger.h"
 #include "gen/passes/Passes.h"
@@ -42,17 +42,17 @@ using namespace llvm;
 
 static cl::opt<signed char> optimizeLevel(
     cl::desc("Setting the optimization level:"), cl::ZeroOrMore,
-    cl::values(
-        clEnumValN(3, "O", "Equivalent to -O3"),
-        clEnumValN(0, "O0", "No optimizations (default)"),
-        clEnumValN(1, "O1", "Simple optimizations"),
-        clEnumValN(2, "O2", "Good optimizations"),
-        clEnumValN(3, "O3", "Aggressive optimizations"),
-        clEnumValN(4, "O4", "Link-time optimization"), // Not implemented yet.
-        clEnumValN(5, "O5", "Link-time optimization"), // Not implemented yet.
-        clEnumValN(-1, "Os", "Like -O2 with extra optimizations for size"),
-        clEnumValN(-2, "Oz", "Like -Os but reduces code size further"),
-        clEnumValEnd),
+    cl::values(clEnumValN(3, "O", "Equivalent to -O3"),
+               clEnumValN(0, "O0", "No optimizations (default)"),
+               clEnumValN(1, "O1", "Simple optimizations"),
+               clEnumValN(2, "O2", "Good optimizations"),
+               clEnumValN(3, "O3", "Aggressive optimizations"),
+               clEnumValN(4, "O4", "Equivalent to -O3"), // Not implemented yet.
+               clEnumValN(5, "O5", "Equivalent to -O3"), // Not implemented yet.
+               clEnumValN(-1, "Os",
+                          "Like -O2 with extra optimizations for size"),
+               clEnumValN(-2, "Oz", "Like -Os but reduces code size further"),
+               clEnumValEnd),
     cl::init(0));
 
 static cl::opt<bool> noVerify("disable-verify",
@@ -202,6 +202,21 @@ static void addThreadSanitizerPass(const PassManagerBuilder &Builder,
   PM.add(createThreadSanitizerPass());
 }
 
+static void addInstrProfilingPass(legacy::PassManagerBase &mpm) {
+#if LDC_WITH_PGO
+  if (global.params.genInstrProf) {
+    InstrProfOptions options;
+    options.NoRedZone = global.params.disableRedZone;
+    options.InstrProfileOutput = global.params.datafileInstrProf;
+#if LDC_LLVM_VER >= 309
+    mpm.add(createInstrProfilingLegacyPass(options));
+#else
+    mpm.add(createInstrProfilingPass(options));
+#endif
+  }
+#endif
+}
+
 /**
  * Adds a set of optimization passes to the given module/function pass
  * managers based on the given optimization and size reduction levels.
@@ -217,7 +232,9 @@ static void addOptimizationPasses(PassManagerBase &mpm,
                                   FunctionPassManager &fpm,
 #endif
                                   unsigned optLevel, unsigned sizeLevel) {
-  fpm.add(createVerifierPass()); // Verify that input is correct
+  if (!noVerify) {
+    fpm.add(createVerifierPass());
+  }
 
   PassManagerBuilder builder;
   builder.OptLevel = optLevel;
@@ -293,6 +310,8 @@ static void addOptimizationPasses(PassManagerBase &mpm,
   builder.addExtension(PassManagerBuilder::EP_OptimizerLast,
                        addStripExternalsPass);
 
+  addInstrProfilingPass(mpm);
+
   builder.populateFunctionPassManager(fpm);
   builder.populateModulePassManager(mpm);
 }
@@ -391,7 +410,9 @@ bool ldc_optimize_module(llvm::Module *M) {
   mpm.run(*M);
 
   // Verify the resulting module.
-  verifyModule(M);
+  if (!noVerify) {
+    verifyModule(M);
+  }
 
   // Report that we run some passes.
   return true;
@@ -399,16 +420,13 @@ bool ldc_optimize_module(llvm::Module *M) {
 
 // Verifies the module.
 void verifyModule(llvm::Module *m) {
-  if (!noVerify) {
-    Logger::println("Verifying module...");
-    LOG_SCOPE;
-    std::string ErrorStr;
-    raw_string_ostream OS(ErrorStr);
-    if (llvm::verifyModule(*m, &OS)) {
-      error(Loc(), "%s", ErrorStr.c_str());
-      fatal();
-    } else {
-      Logger::println("Verification passed!");
-    }
+  Logger::println("Verifying module...");
+  LOG_SCOPE;
+  std::string ErrorStr;
+  raw_string_ostream OS(ErrorStr);
+  if (llvm::verifyModule(*m, &OS)) {
+    error(Loc(), "%s", ErrorStr.c_str());
+    fatal();
   }
+  Logger::println("Verification passed!");
 }
