@@ -10,11 +10,6 @@
 #include "gen/runtime.h"
 #include "aggregate.h"
 #include "dsymbol.h"
-#include "lexer.h"
-#include "mars.h"
-#include "module.h"
-#include "mtype.h"
-#include "root.h"
 #include "gen/abi.h"
 #include "gen/attributes.h"
 #include "gen/functions.h"
@@ -23,13 +18,20 @@
 #include "gen/llvmhelpers.h"
 #include "gen/logger.h"
 #include "gen/tollvm.h"
+#include "ir/irfunction.h"
 #include "ir/irtype.h"
 #include "ir/irtypefunction.h"
+#include "ldcbindings.h"
+#include "mars.h"
+#include "module.h"
+#include "mtype.h"
+#include "root.h"
+#include "tokens.h"
 #include "llvm/Bitcode/ReaderWriter.h"
+#include "llvm/IR/Attributes.h"
+#include "llvm/IR/Module.h"
 #include "llvm/Support/CommandLine.h"
 #include "llvm/Support/MemoryBuffer.h"
-#include "llvm/IR/Module.h"
-#include "llvm/IR/Attributes.h"
 
 #include <algorithm>
 
@@ -183,9 +185,9 @@ static Type *rt_dg1() {
     return dg_t;
 
   auto params = new Parameters();
-  params->push(new Parameter(0, Type::tvoidptr, nullptr, nullptr));
-  auto fty = new TypeFunction(params, Type::tint32, 0, LINKd);
-  dg_t = new TypeDelegate(fty);
+  params->push(Parameter::create(0, Type::tvoidptr, nullptr, nullptr));
+  auto fty = TypeFunction::create(params, Type::tint32, 0, LINKd);
+  dg_t = createTypeDelegate(fty);
   return dg_t;
 }
 
@@ -196,10 +198,10 @@ static Type *rt_dg2() {
     return dg2_t;
 
   auto params = new Parameters();
-  params->push(new Parameter(0, Type::tvoidptr, nullptr, nullptr));
-  params->push(new Parameter(0, Type::tvoidptr, nullptr, nullptr));
-  auto fty = new TypeFunction(params, Type::tint32, 0, LINKd);
-  dg2_t = new TypeDelegate(fty);
+  params->push(Parameter::create(0, Type::tvoidptr, nullptr, nullptr));
+  params->push(Parameter::create(0, Type::tvoidptr, nullptr, nullptr));
+  auto fty = TypeFunction::create(params, Type::tint32, 0, LINKd);
+  dg2_t = createTypeDelegate(fty);
   return dg2_t;
 }
 
@@ -213,13 +215,6 @@ template <typename DECL> static void ensureDecl(DECL *decl, const char *msg) {
   }
 }
 
-// We need a vector type with initializer_list support:
-#if LDC_LLVM_VER >= 307
-template <class T> using ArrayParam = llvm::ArrayRef<T>;
-#else
-template <class T> using ArrayParam = std::vector<T>;
-#endif
-
 // Parameters fnames are assumed to be already mangled!
 static void createFwdDecl(LINK linkage, Type *returntype,
                           ArrayParam<llvm::StringRef> fnames,
@@ -232,11 +227,11 @@ static void createFwdDecl(LINK linkage, Type *returntype,
     params = new Parameters();
     for (size_t i = 0, e = paramtypes.size(); i < e; ++i) {
       StorageClass stc = paramsSTC.empty() ? 0 : paramsSTC[i];
-      params->push(new Parameter(stc, paramtypes[i], nullptr, nullptr));
+      params->push(Parameter::create(stc, paramtypes[i], nullptr, nullptr));
     }
   }
   int varargs = 0;
-  auto dty = new TypeFunction(params, returntype, varargs, linkage);
+  auto dty = TypeFunction::create(params, returntype, varargs, linkage);
 
   // the call to DtoType performs many actions such as rewriting the function
   // type and storing it in dty
@@ -252,6 +247,13 @@ static void createFwdDecl(LINK linkage, Type *returntype,
 
     fn->setAttributes(attrs);
 
+    // On x86_64, always set 'uwtable' for System V ABI compatibility.
+    // FIXME: Move to better place (abi-x86-64.cpp?)
+    // NOTE: There are several occurances if this line.
+    if (global.params.targetTriple->getArch() == llvm::Triple::x86_64) {
+      fn->addFnAttr(LLAttribute::UWTable);
+    }
+
     fn->setCallingConv(gABI->callingConv(fn->getFunctionType(), linkage));
   }
 }
@@ -296,10 +298,18 @@ static void buildRuntimeModule() {
   //////////////////////////////////////////////////////////////////////////////
 
   // Construct some attribute lists used below (possibly multiple times)
-  AttrSet NoAttrs, Attr_NoAlias(NoAttrs, 0, llvm::Attribute::NoAlias),
-      Attr_NoUnwind(NoAttrs, ~0U, llvm::Attribute::NoUnwind),
-      Attr_ReadOnly(NoAttrs, ~0U, llvm::Attribute::ReadOnly),
-      Attr_ReadOnly_NoUnwind(Attr_ReadOnly, ~0U, llvm::Attribute::NoUnwind),
+  AttrSet NoAttrs, Attr_NoAlias(NoAttrs, llvm::AttributeSet::ReturnIndex,
+                                llvm::Attribute::NoAlias),
+      Attr_NoUnwind(NoAttrs, llvm::AttributeSet::FunctionIndex,
+                    llvm::Attribute::NoUnwind),
+      Attr_ReadOnly(NoAttrs, llvm::AttributeSet::FunctionIndex,
+                    llvm::Attribute::ReadOnly),
+      Attr_Cold(NoAttrs, llvm::AttributeSet::FunctionIndex,
+                llvm::Attribute::Cold),
+      Attr_Cold_NoReturn(Attr_Cold, llvm::AttributeSet::FunctionIndex,
+                         llvm::Attribute::NoReturn),
+      Attr_ReadOnly_NoUnwind(Attr_ReadOnly, llvm::AttributeSet::FunctionIndex,
+                             llvm::Attribute::NoUnwind),
       Attr_ReadOnly_1_NoCapture(Attr_ReadOnly, 1, llvm::Attribute::NoCapture),
       Attr_ReadOnly_1_3_NoCapture(Attr_ReadOnly_1_NoCapture, 3,
                                   llvm::Attribute::NoCapture),
@@ -319,17 +329,18 @@ static void buildRuntimeModule() {
   // void _d_assert(string file, uint line)
   // void _d_arraybounds(string file, uint line)
   createFwdDecl(LINKc, Type::tvoid, {"_d_assert", "_d_arraybounds"},
-                {stringTy, uintTy});
+                {stringTy, uintTy}, {}, Attr_Cold_NoReturn);
 
   // void _d_assert_msg(string msg, string file, uint line)
-  createFwdDecl(LINKc, voidTy, {"_d_assert_msg"}, {stringTy, stringTy, uintTy});
+  createFwdDecl(LINKc, voidTy, {"_d_assert_msg"}, {stringTy, stringTy, uintTy},
+                {}, Attr_Cold_NoReturn);
 
   // void _d_assertm(immutable(ModuleInfo)* m, uint line)
   // void _d_array_bounds(immutable(ModuleInfo)* m, uint line)
   // void _d_switch_error(immutable(ModuleInfo)* m, uint line)
-  createFwdDecl(LINKc, voidTy,
-                {"_d_assertm", "_d_array_bounds", "_d_switch_error"},
-                {moduleInfoPtrTy, uintTy}, {STCimmutable, 0});
+  createFwdDecl(
+      LINKc, voidTy, {"_d_assertm", "_d_array_bounds", "_d_switch_error"},
+      {moduleInfoPtrTy, uintTy}, {STCimmutable, 0}, Attr_Cold_NoReturn);
 
   //////////////////////////////////////////////////////////////////////////////
   //////////////////////////////////////////////////////////////////////////////
@@ -618,12 +629,14 @@ static void buildRuntimeModule() {
 
   // int _d_eh_personality(...)
   {
-    if (global.params.targetTriple.isWindowsMSVCEnvironment()) {
+    if (global.params.targetTriple->isWindowsMSVCEnvironment()) {
+      const char *fname =
+          useMSVCEH() ? "__CxxFrameHandler3" : "_d_eh_personality";
       // (ptr ExceptionRecord, ptr EstablisherFrame, ptr ContextRecord,
       //  ptr DispatcherContext)
-      createFwdDecl(LINKc, intTy, {"_d_eh_personality"},
+      createFwdDecl(LINKc, intTy, {fname},
                     {voidPtrTy, voidPtrTy, voidPtrTy, voidPtrTy});
-    } else if (global.params.targetTriple.getArch() == llvm::Triple::arm) {
+    } else if (global.params.targetTriple->getArch() == llvm::Triple::arm) {
       // (int state, ptr ucb, ptr context)
       createFwdDecl(LINKc, intTy, {"_d_eh_personality"},
                     {intTy, voidPtrTy, voidPtrTy});
@@ -634,12 +647,25 @@ static void buildRuntimeModule() {
     }
   }
 
-  // void _d_eh_resume_unwind(ptr)
-  createFwdDecl(LINKc, voidTy, {"_d_eh_resume_unwind"}, {voidPtrTy});
+  if (useMSVCEH()) {
+    // _d_enter_cleanup(ptr frame)
+    createFwdDecl(LINKc, boolTy, {"_d_enter_cleanup"}, {voidPtrTy});
 
-  // Object _d_eh_enter_catch(ptr)
-  createFwdDecl(LINKc, objectTy, {"_d_eh_enter_catch"}, {voidPtrTy}, {},
-                Attr_NoUnwind);
+    // _d_leave_cleanup(ptr frame)
+    createFwdDecl(LINKc, voidTy, {"_d_leave_cleanup"}, {voidPtrTy});
+
+    // Object _d_eh_enter_catch(ptr exception, ClassInfo catchType)
+    createFwdDecl(LINKc, objectTy, {"_d_eh_enter_catch"},
+                  {voidPtrTy, classInfoTy}, {});
+  } else {
+
+    // void _d_eh_resume_unwind(ptr)
+    createFwdDecl(LINKc, voidTy, {"_d_eh_resume_unwind"}, {voidPtrTy});
+
+    // Object _d_eh_enter_catch(ptr)
+    createFwdDecl(LINKc, objectTy, {"_d_eh_enter_catch"}, {voidPtrTy}, {},
+                  Attr_NoUnwind);
+  }
 
   //////////////////////////////////////////////////////////////////////////////
   //////////////////////////////////////////////////////////////////////////////
@@ -648,7 +674,7 @@ static void buildRuntimeModule() {
   // void invariant._d_invariant(Object o)
   createFwdDecl(
       LINKd, voidTy,
-      {gABI->mangleForLLVM("_D9invariant12_d_invariantFC6ObjectZv", LINKd)},
+      {gABI->mangleFunctionForLLVM("_D9invariant12_d_invariantFC6ObjectZv", LINKd)},
       {objectTy});
 
   // void _d_dso_registry(CompilerDSOData* data)
@@ -680,4 +706,43 @@ static void buildRuntimeModule() {
     createFwdDecl(LINKc, voidTy, {"_d_cover_register2"},
                   {stringTy, sizeTy->arrayOf(), uintTy->arrayOf(), ubyteTy});
   }
+
+  if (global.params.hasObjectiveC) {
+    assert(global.params.targetTriple->isOSDarwin());
+
+    // The types of these functions don't really matter because they are always
+    // bitcast to correct signature before calling.
+    Type* objectPtrTy = voidPtrTy;
+    Type* selectorPtrTy = voidPtrTy;
+    Type* realTy = Type::tfloat80;
+
+    // id objc_msgSend(id self, SEL op, ...)
+    // Function called early and/or often, so lazy binding isn't worthwhile.
+    createFwdDecl(LINKc, objectPtrTy, {"objc_msgSend"},
+                  {objectPtrTy, selectorPtrTy}, {},
+                  AttrSet(NoAttrs, ~0U, llvm::Attribute::NonLazyBind));
+
+    switch (global.params.targetTriple->getArch()) {
+    case llvm::Triple::x86_64:
+      // creal objc_msgSend_fp2ret(id self, SEL op, ...)
+      createFwdDecl(LINKc, Type::tcomplex80, {"objc_msgSend_fp2ret"},
+                    {objectPtrTy, selectorPtrTy});
+      // fall-thru
+    case llvm::Triple::x86:
+      // x86_64 real return only,  x86 float, double, real return
+      // real objc_msgSend_fpret(id self, SEL op, ...)
+      createFwdDecl(LINKc, realTy, {"objc_msgSend_fpret"},
+                    {objectPtrTy, selectorPtrTy});
+      // fall-thru
+    case llvm::Triple::arm:
+    case llvm::Triple::thumb:
+      // used when return value is aggregate via a hidden sret arg
+      // void objc_msgSend_stret(T *sret_arg, id self, SEL op, ...)
+      createFwdDecl(LINKc, voidTy, {"objc_msgSend_stret"},
+                    {objectPtrTy, selectorPtrTy});
+      break;
+    default:
+      break;
+    }
+  }
 }
