@@ -13,6 +13,7 @@
 #include "target.h"
 #include "pragma.h"
 #include "gen/abi.h"
+#include "gen/classes.h"
 #include "gen/dvalue.h"
 #include "gen/functions.h"
 #include "gen/irstate.h"
@@ -20,7 +21,9 @@
 #include "gen/llvmhelpers.h"
 #include "gen/logger.h"
 #include "gen/nested.h"
+#include "gen/objcgen.h"
 #include "gen/tollvm.h"
+#include "gen/runtime.h"
 #include "ir/irfunction.h"
 #include "ir/irtype.h"
 
@@ -33,14 +36,14 @@ IrFuncTy &DtoIrTypeFunction(DValue *fnval) {
     }
   }
 
-  Type *type = stripModifiers(fnval->getType()->toBasetype());
+  Type *type = stripModifiers(fnval->type->toBasetype());
   DtoType(type);
   assert(type->ctype);
   return type->ctype->getIrFuncTy();
 }
 
 TypeFunction *DtoTypeFunction(DValue *fnval) {
-  Type *type = fnval->getType()->toBasetype();
+  Type *type = fnval->type->toBasetype();
   if (type->ty == Tfunction) {
     return static_cast<TypeFunction *>(type);
   }
@@ -69,17 +72,17 @@ TypeFunction *DtoTypeFunction(DValue *fnval) {
 ////////////////////////////////////////////////////////////////////////////////
 
 LLValue *DtoCallableValue(DValue *fn) {
-  Type *type = fn->getType()->toBasetype();
+  Type *type = fn->type->toBasetype();
   if (type->ty == Tfunction) {
-    return fn->getRVal();
+    return DtoRVal(fn);
   }
   if (type->ty == Tdelegate) {
     if (fn->isLVal()) {
-      LLValue *dg = fn->getLVal();
+      LLValue *dg = DtoLVal(fn);
       LLValue *funcptr = DtoGEPi(dg, 0, 1);
       return DtoLoad(funcptr, ".funcptr");
     }
-    LLValue *dg = fn->getRVal();
+    LLValue *dg = DtoRVal(fn);
     assert(isaStruct(dg));
     return gIR->ir->CreateExtractValue(dg, 1, ".funcptr");
   }
@@ -104,7 +107,7 @@ LLFunctionType *DtoExtractFunctionType(LLType *type) {
 ////////////////////////////////////////////////////////////////////////////////
 
 static void addExplicitArguments(std::vector<LLValue *> &args, AttrSet &attrs,
-                                 IrFuncTy &irFty, LLFunctionType *callableTy,
+                                 IrFuncTy &irFty, LLFunctionType *calleeType,
                                  const std::vector<DValue *> &argvals,
                                  int numFormalParams) {
   // Number of arguments added to the LLVM type that are implicit on the
@@ -121,7 +124,7 @@ static void addExplicitArguments(std::vector<LLValue *> &args, AttrSet &attrs,
   // construct and initialize an IrFuncTyArg object for each vararg
   std::vector<IrFuncTyArg *> optionalIrArgs;
   for (size_t i = numFormalParams; i < explicitDArgCount; i++) {
-    Type *argType = argvals[i]->getType();
+    Type *argType = argvals[i]->type;
     bool passByVal = gABI->passByVal(argType);
 
     AttrBuilder initialAttrs;
@@ -154,7 +157,7 @@ static void addExplicitArguments(std::vector<LLValue *> &args, AttrSet &attrs,
     }
 
     DValue *const argval = argvals[irArg->parametersIdx];
-    Type *const argType = argval->getType();
+    Type *const argType = argval->type;
 
     llvm::Value *llVal = nullptr;
     if (isVararg) {
@@ -166,29 +169,29 @@ static void addExplicitArguments(std::vector<LLValue *> &args, AttrSet &attrs,
     const size_t llArgIdx =
         implicitLLArgCount +
         (irFty.reverseParams ? explicitLLArgCount - i - 1 : i);
-    llvm::Type *const callableArgType =
-        (isVararg ? nullptr : callableTy->getParamType(llArgIdx));
+    llvm::Type *const paramType =
+        (isVararg ? nullptr : calleeType->getParamType(llArgIdx));
 
     // Hack around LDC assuming structs and static arrays are in memory:
     // If the function wants a struct, and the argument value is a
     // pointer to a struct, load from it before passing it in.
     if (isaPointer(llVal) && DtoIsInMemoryOnly(argType) &&
-        ((!isVararg && !isaPointer(callableArgType)) ||
+        ((!isVararg && !isaPointer(paramType)) ||
          (isVararg && !irArg->byref && !irArg->isByVal()))) {
       Logger::println("Loading struct type for function argument");
       llVal = DtoLoad(llVal);
     }
 
     // parameter type mismatch, this is hard to get rid of
-    if (!isVararg && llVal->getType() != callableArgType) {
+    if (!isVararg && llVal->getType() != paramType) {
       IF_LOG {
         Logger::cout() << "arg:     " << *llVal << '\n';
-        Logger::cout() << "expects: " << *callableArgType << '\n';
+        Logger::cout() << "expects: " << *paramType << '\n';
       }
       if (isaStruct(llVal)) {
-        llVal = DtoAggrPaint(llVal, callableArgType);
+        llVal = DtoAggrPaint(llVal, paramType);
       } else {
-        llVal = DtoBitCast(llVal, callableArgType);
+        llVal = DtoBitCast(llVal, paramType);
       }
     }
 
@@ -261,17 +264,16 @@ bool DtoLowerMagicIntrinsic(IRState *p, FuncDeclaration *fndecl, CallExp *e,
       e->error("va_start instruction expects 1 (or 2) arguments");
       fatal();
     }
-    LLValue *pAp = toElem((*e->arguments)[0])->getLVal(); // va_list*
+    DLValue *ap = toElem((*e->arguments)[0])->isLVal(); // va_list
+    assert(ap);
     // variadic extern(D) function with implicit _argptr?
-    if (LLValue *pArgptr = p->func()->_argptr) {
-      DtoMemCpy(pAp, pArgptr); // ap = _argptr
-      result = new DImValue(e->type, pAp);
+    if (LLValue *argptrMem = p->func()->_argptr) {
+      DtoMemCpy(DtoLVal(ap), argptrMem); // ap = _argptr
     } else {
-      LLValue *vaStartArg = gABI->prepareVaStart(pAp);
-      result =
-          new DImValue(e->type, p->ir->CreateCall(GET_INTRINSIC_DECL(vastart),
-                                                  vaStartArg, ""));
+      LLValue *llAp = gABI->prepareVaStart(ap);
+      p->ir->CreateCall(GET_INTRINSIC_DECL(vastart), llAp, "");
     }
+    result = nullptr;
     return true;
   }
 
@@ -281,10 +283,11 @@ bool DtoLowerMagicIntrinsic(IRState *p, FuncDeclaration *fndecl, CallExp *e,
       e->error("va_copy instruction expects 2 arguments");
       fatal();
     }
-    LLValue *pDest = toElem((*e->arguments)[0])->getLVal(); // va_list*
-    LLValue *src = toElem((*e->arguments)[1])->getRVal();   // va_list
-    gABI->vaCopy(pDest, src);
-    result = new DVarValue(e->type, pDest);
+    DLValue *dest = toElem((*e->arguments)[0])->isLVal(); // va_list
+    assert(dest);
+    DValue *src = toElem((*e->arguments)[1]);             // va_list
+    gABI->vaCopy(dest, src);
+    result = nullptr;
     return true;
   }
 
@@ -294,13 +297,15 @@ bool DtoLowerMagicIntrinsic(IRState *p, FuncDeclaration *fndecl, CallExp *e,
       e->error("va_arg instruction expects 1 argument");
       fatal();
     }
-    LLValue *pAp = toElem((*e->arguments)[0])->getLVal(); // va_list*
-    LLValue *vaArgArg = gABI->prepareVaArg(pAp);
-    LLType *llType = DtoType(e->type);
     if (DtoIsInMemoryOnly(e->type)) {
-      llType = getPtrToType(llType);
+      e->error("va_arg instruction does not support structs and static arrays");
+      fatal();
     }
-    result = new DImValue(e->type, p->ir->CreateVAArg(vaArgArg, llType));
+    DLValue *ap = toElem((*e->arguments)[0])->isLVal(); // va_list
+    assert(ap);
+    LLValue *llAp = gABI->prepareVaArg(ap);
+    LLType *llType = DtoType(e->type);
+    result = new DImValue(e->type, p->ir->CreateVAArg(llAp, llType));
     return true;
   }
 
@@ -312,12 +317,12 @@ bool DtoLowerMagicIntrinsic(IRState *p, FuncDeclaration *fndecl, CallExp *e,
     }
     Expression *exp = (*e->arguments)[0];
     DValue *expv = toElem(exp);
-    if (expv->getType()->toBasetype()->ty != Tint32) {
+    if (expv->type->toBasetype()->ty != Tint32) {
       expv = DtoCast(e->loc, expv, Type::tint32);
     }
     result = new DImValue(e->type,
                           p->ir->CreateAlloca(LLType::getInt8Ty(p->context()),
-                                              expv->getRVal(), ".alloca"));
+                                              DtoRVal(expv), ".alloca"));
     return true;
   }
 
@@ -340,24 +345,27 @@ bool DtoLowerMagicIntrinsic(IRState *p, FuncDeclaration *fndecl, CallExp *e,
     Expression *exp1 = (*e->arguments)[0];
     Expression *exp2 = (*e->arguments)[1];
     int atomicOrdering = (*e->arguments)[2]->toInteger();
-    LLValue *val = toElem(exp1)->getRVal();
-    LLValue *ptr = toElem(exp2)->getRVal();
 
-    if (!val->getType()->isIntegerTy()) {
-      llvm::PointerType *v = isaPointer(val->getType());
-      if (v && v->getContainedType(0)->isStructTy()) {
-        switch (const size_t N = getTypeBitSize(v->getContainedType(0))) {
+    DValue *dval = toElem(exp1);
+    LLValue *ptr = DtoRVal(exp2);
+    LLType *pointeeType = ptr->getType()->getContainedType(0);
+
+    LLValue *val = nullptr;
+    if (!pointeeType->isIntegerTy()) {
+      if (pointeeType->isStructTy()) {
+        val = DtoLVal(dval);
+        switch (size_t N = getTypeBitSize(pointeeType)) {
         case 8:
         case 16:
         case 32:
         case 64:
-        case 128:
-          val = DtoLoad(
-              DtoBitCast(val, llvm::Type::getIntNPtrTy(
-                                  gIR->context(), static_cast<unsigned>(N))));
-          ptr = DtoBitCast(ptr, llvm::Type::getIntNPtrTy(
-                                    gIR->context(), static_cast<unsigned>(N)));
+        case 128: {
+          LLType *intPtrType =
+              LLType::getIntNPtrTy(gIR->context(), static_cast<unsigned>(N));
+          val = DtoLoad(DtoBitCast(val, intPtrType));
+          ptr = DtoBitCast(ptr, intPtrType);
           break;
+        }
         default:
           goto errorStore;
         }
@@ -367,6 +375,8 @@ bool DtoLowerMagicIntrinsic(IRState *p, FuncDeclaration *fndecl, CallExp *e,
                  exp1->type->toChars());
         fatal();
       }
+    } else {
+      val = DtoRVal(dval);
     }
 
     llvm::StoreInst *ret = p->ir->CreateStore(val, ptr);
@@ -385,20 +395,20 @@ bool DtoLowerMagicIntrinsic(IRState *p, FuncDeclaration *fndecl, CallExp *e,
     Expression *exp = (*e->arguments)[0];
     int atomicOrdering = (*e->arguments)[1]->toInteger();
 
-    LLValue *ptr = toElem(exp)->getRVal();
-    LLType *ptrTy = ptr->getType()->getContainedType(0);
+    LLValue *ptr = DtoRVal(exp);
+    LLType *pointeeType = ptr->getType()->getContainedType(0);
     Type *retType = exp->type->nextOf();
 
-    if (!ptrTy->isIntegerTy()) {
-      if (ptrTy->isStructTy()) {
-        switch (const size_t N = getTypeBitSize(ptrTy)) {
+    if (!pointeeType->isIntegerTy()) {
+      if (pointeeType->isStructTy()) {
+        switch (const size_t N = getTypeBitSize(pointeeType)) {
         case 8:
         case 16:
         case 32:
         case 64:
         case 128:
-          ptr = DtoBitCast(ptr, llvm::Type::getIntNPtrTy(
-                                    gIR->context(), static_cast<unsigned>(N)));
+          ptr = DtoBitCast(ptr, LLType::getIntNPtrTy(gIR->context(),
+                                                     static_cast<unsigned>(N)));
           break;
         default:
           goto errorLoad;
@@ -415,10 +425,12 @@ bool DtoLowerMagicIntrinsic(IRState *p, FuncDeclaration *fndecl, CallExp *e,
     load->setAlignment(getTypeAllocSize(load->getType()));
     load->setAtomic(llvm::AtomicOrdering(atomicOrdering));
     llvm::Value *val = load;
-    if (val->getType() != ptrTy) {
+    if (val->getType() != pointeeType) {
       val = DtoAllocaDump(val, retType);
+      result = new DLValue(retType, val);
+    } else {
+      result = new DImValue(retType, val);
     }
-    result = new DImValue(retType, val);
     return true;
   }
 
@@ -431,30 +443,29 @@ bool DtoLowerMagicIntrinsic(IRState *p, FuncDeclaration *fndecl, CallExp *e,
     Expression *exp1 = (*e->arguments)[0];
     Expression *exp2 = (*e->arguments)[1];
     Expression *exp3 = (*e->arguments)[2];
-    int atomicOrdering = (*e->arguments)[3]->toInteger();
-    LLValue *ptr = toElem(exp1)->getRVal();
-    LLValue *cmp = toElem(exp2)->getRVal();
-    LLValue *val = toElem(exp3)->getRVal();
-    LLType *retTy = val->getType();
+    auto atomicOrdering = llvm::AtomicOrdering((*e->arguments)[3]->toInteger());
+    LLValue *ptr = DtoRVal(exp1);
+    LLType *pointeeType = ptr->getType()->getContainedType(0);
+    DValue *dcmp = toElem(exp2);
+    DValue *dval = toElem(exp3);
 
-    if (!cmp->getType()->isIntegerTy()) {
-      llvm::PointerType *v = isaPointer(cmp->getType());
-      if (v && v->getContainedType(0)->isStructTy()) {
-        switch (const size_t N = getTypeBitSize(v->getContainedType(0))) {
+    LLValue *cmp = nullptr;
+    LLValue *val = nullptr;
+    if (!pointeeType->isIntegerTy()) {
+      if (pointeeType->isStructTy()) {
+        switch (const size_t N = getTypeBitSize(pointeeType)) {
         case 8:
         case 16:
         case 32:
         case 64:
-        case 128:
-          ptr = DtoBitCast(ptr, llvm::Type::getIntNPtrTy(
-                                    gIR->context(), static_cast<unsigned>(N)));
-          cmp = DtoLoad(
-              DtoBitCast(cmp, llvm::Type::getIntNPtrTy(
-                                  gIR->context(), static_cast<unsigned>(N))));
-          val = DtoLoad(
-              DtoBitCast(val, llvm::Type::getIntNPtrTy(
-                                  gIR->context(), static_cast<unsigned>(N))));
+        case 128: {
+          LLType *intPtrType =
+              LLType::getIntNPtrTy(gIR->context(), static_cast<unsigned>(N));
+          ptr = DtoBitCast(ptr, intPtrType);
+          cmp = DtoLoad(DtoBitCast(DtoLVal(dcmp), intPtrType));
+          val = DtoLoad(DtoBitCast(DtoLVal(dval), intPtrType));
           break;
+        }
         default:
           goto errorCmpxchg;
         }
@@ -464,17 +475,21 @@ bool DtoLowerMagicIntrinsic(IRState *p, FuncDeclaration *fndecl, CallExp *e,
                  exp2->type->toChars());
         fatal();
       }
+    } else {
+      cmp = DtoRVal(dcmp);
+      val = DtoRVal(dval);
     }
-    LLValue *ret = p->ir->CreateAtomicCmpXchg(
-        ptr, cmp, val, llvm::AtomicOrdering(atomicOrdering),
-        llvm::AtomicOrdering(atomicOrdering));
+
+    LLValue *ret = p->ir->CreateAtomicCmpXchg(ptr, cmp, val, atomicOrdering,
+                                              atomicOrdering);
     // Use the same quickfix as for dragonegg - see r210956
     ret = p->ir->CreateExtractValue(ret, 0);
-    llvm::Value *retVal = ret;
-    if (retVal->getType() != retTy) {
-      retVal = DtoAllocaDump(retVal, exp3->type);
+    if (ret->getType() != pointeeType) {
+      ret = DtoAllocaDump(ret, exp3->type);
+      result = new DLValue(exp3->type, ret);
+    } else {
+      result = new DImValue(exp3->type, ret);
     }
-    result = new DImValue(exp3->type, retVal);
     return true;
   }
 
@@ -485,6 +500,7 @@ bool DtoLowerMagicIntrinsic(IRState *p, FuncDeclaration *fndecl, CallExp *e,
       fatal();
     }
 
+    assert(fndecl->intrinsicName);
     static const char *ops[] = {"xchg", "add", "sub", "and",  "nand", "or",
                                 "xor",  "max", "min", "umax", "umin", nullptr};
 
@@ -492,10 +508,10 @@ bool DtoLowerMagicIntrinsic(IRState *p, FuncDeclaration *fndecl, CallExp *e,
     for (;; ++op) {
       if (ops[op] == nullptr) {
         e->error("unknown atomic_rmw operation %s",
-                 fndecl->intrinsicName.c_str());
+                 fndecl->intrinsicName);
         fatal();
       }
-      if (fndecl->intrinsicName == ops[op]) {
+      if (strcmp(fndecl->intrinsicName, ops[op]) == 0) {
         break;
       }
     }
@@ -503,8 +519,8 @@ bool DtoLowerMagicIntrinsic(IRState *p, FuncDeclaration *fndecl, CallExp *e,
     Expression *exp1 = (*e->arguments)[0];
     Expression *exp2 = (*e->arguments)[1];
     int atomicOrdering = (*e->arguments)[2]->toInteger();
-    LLValue *ptr = toElem(exp1)->getRVal();
-    LLValue *val = toElem(exp2)->getRVal();
+    LLValue *ptr = DtoRVal(exp1);
+    LLValue *val = DtoRVal(exp2);
     LLValue *ret =
         p->ir->CreateAtomicRMW(llvm::AtomicRMWInst::BinOp(op), ptr, val,
                                llvm::AtomicOrdering(atomicOrdering));
@@ -524,8 +540,8 @@ bool DtoLowerMagicIntrinsic(IRState *p, FuncDeclaration *fndecl, CallExp *e,
 
     Expression *exp1 = (*e->arguments)[0];
     Expression *exp2 = (*e->arguments)[1];
-    LLValue *ptr = toElem(exp1)->getRVal();
-    LLValue *bitnum = toElem(exp2)->getRVal();
+    LLValue *ptr = DtoRVal(exp1);
+    LLValue *bitnum = DtoRVal(exp2);
 
     unsigned bitmask = DtoSize_t()->getBitWidth() - 1;
     assert(bitmask == 31 || bitmask == 63);
@@ -579,7 +595,7 @@ bool DtoLowerMagicIntrinsic(IRState *p, FuncDeclaration *fndecl, CallExp *e,
     // TODO: Check types
 
     Expression *exp1 = (*e->arguments)[0];
-    LLValue *ptr = toElem(exp1)->getRVal();
+    LLValue *ptr = DtoRVal(exp1);
     result = new DImValue(exp1->type, DtoVolatileLoad(ptr));
     return true;
   }
@@ -593,8 +609,8 @@ bool DtoLowerMagicIntrinsic(IRState *p, FuncDeclaration *fndecl, CallExp *e,
 
     Expression *exp1 = (*e->arguments)[0];
     Expression *exp2 = (*e->arguments)[1];
-    LLValue *ptr = toElem(exp1)->getRVal();
-    LLValue *val = toElem(exp2)->getRVal();
+    LLValue *ptr = DtoRVal(exp1);
+    LLValue *val = DtoRVal(exp2);
     DtoVolatileStore(val, ptr);
     return true;
   }
@@ -606,6 +622,8 @@ bool DtoLowerMagicIntrinsic(IRState *p, FuncDeclaration *fndecl, CallExp *e,
 
 class ImplicitArgumentsBuilder {
 public:
+  bool hasObjcSelector = false;
+
   ImplicitArgumentsBuilder(std::vector<LLValue *> &args, AttrSet &attrs,
                            Loc &loc, DValue *fnval,
                            LLFunctionType *llCalleeType, Expressions *arguments,
@@ -613,7 +631,7 @@ public:
       : args(args), attrs(attrs), loc(loc), fnval(fnval), arguments(arguments),
         resulttype(resulttype), retvar(retvar),
         // computed:
-        calleeType(fnval->getType()), dfnval(fnval->isFunc()),
+        calleeType(fnval->type), dfnval(fnval->isFunc()),
         irFty(DtoIrTypeFunction(fnval)), tf(DtoTypeFunction(fnval)),
         llArgTypesBegin(llCalleeType->param_begin()) {}
 
@@ -677,9 +695,8 @@ private:
     bool delegatecall = (calleeType->toBasetype()->ty == Tdelegate);
     bool nestedcall = irFty.arg_nest;
 
-    if (!thiscall && !delegatecall && !nestedcall) {
+    if (!thiscall && !delegatecall && !nestedcall)
       return;
-    }
 
     size_t index = args.size();
     LLType *llArgType = *(llArgTypesBegin + index);
@@ -691,8 +708,19 @@ private:
       // __require/__ensure as the needed parameters are passed
       // explicitly, while in D1, the normal nested function handling
       // mechanisms are used)
-      LLValue *thisarg =
-          DtoBitCast(DtoLoad(gIR->func()->thisArg), getVoidPtrType());
+      LLValue *thisptr = DtoLoad(gIR->func()->thisArg);
+      if (auto parentfd = dfnval->func->parent->isFuncDeclaration()) {
+        if (auto iface = parentfd->parent->isInterfaceDeclaration()) {
+          // an interface contract expects the interface pointer, not the
+          //  class pointer
+          Type *thistype = gIR->func()->decl->vthis->type;
+          if (thistype != iface->type) {
+            DImValue *dthis = new DImValue(thistype, thisptr);
+            thisptr = DtoRVal(DtoCastClass(loc, dthis, iface->type));
+          }
+        }
+      }
+      LLValue *thisarg = DtoBitCast(thisptr, getVoidPtrType());
       args.push_back(thisarg);
     } else if (thiscall && dfnval && dfnval->vthis) {
       // ... or a normal 'this' argument
@@ -702,9 +730,9 @@ private:
       // ... or a delegate context arg
       LLValue *ctxarg;
       if (fnval->isLVal()) {
-        ctxarg = DtoLoad(DtoGEPi(fnval->getLVal(), 0, 0), ".ptr");
+        ctxarg = DtoLoad(DtoGEPi(DtoLVal(fnval), 0, 0), ".ptr");
       } else {
-        ctxarg = gIR->ir->CreateExtractValue(fnval->getRVal(), 0, ".ptr");
+        ctxarg = gIR->ir->CreateExtractValue(DtoRVal(fnval), 0, ".ptr");
       }
       ctxarg = DtoBitCast(ctxarg, llArgType);
       args.push_back(ctxarg);
@@ -728,6 +756,14 @@ private:
     } else if (irFty.arg_nest) {
       attrs.add(index + 1, irFty.arg_nest->attrs);
     }
+
+    if (irFty.arg_objcSelector && dfnval) {
+      if (auto sel = dfnval->func->objc.selector) {
+        LLGlobalVariable* selptr = objc_getMethVarRef(*sel);
+        args.push_back(DtoBitCast(DtoLoad(selptr), getVoidPtrType()));
+        hasObjcSelector = true;
+      }
+    }
   }
 
   // D vararg functions need a "TypeInfo[] _arguments" argument.
@@ -755,7 +791,7 @@ DValue *DtoCallFunction(Loc &loc, Type *resulttype, DValue *fnval,
   LOG_SCOPE
 
   // make sure the D callee type has been processed
-  DtoType(fnval->getType());
+  DtoType(fnval->type);
 
   // get func value if any
   DFuncValue *dfnval = fnval->isFunc();
@@ -771,7 +807,7 @@ DValue *DtoCallFunction(Loc &loc, Type *resulttype, DValue *fnval,
   }
 
   // get callee llvm value
-  LLValue *const callable = DtoCallableValue(fnval);
+  LLValue *callable = DtoCallableValue(fnval);
   LLFunctionType *const callableTy =
       DtoExtractFunctionType(callable->getType());
   assert(callableTy);
@@ -840,6 +876,14 @@ DValue *DtoCallFunction(Loc &loc, Type *resulttype, DValue *fnval,
   addExplicitArguments(args, attrs, irFty, callableTy, argvals,
                        numFormalParams);
 
+  if (iab.hasObjcSelector) {
+    // Use runtime msgSend function bitcasted as original call
+    const char *msgSend = gABI->objcMsgSendFunc(resulttype, irFty);
+    LLType *t = callable->getType();
+    callable = getRuntimeFunction(loc, gIR->module, msgSend);
+    callable = DtoBitCast(callable, t);
+  }
+
   // call the function
   LLCallSite call = gIR->func()->scopes->callOrInvoke(callable, args);
 
@@ -849,47 +893,21 @@ DValue *DtoCallFunction(Loc &loc, Type *resulttype, DValue *fnval,
                                                                         : 0);
   LLValue *retllval =
       (irFty.arg_sret ? args[sretArgIndex] : call.getInstruction());
+  bool retValIsLVal = (tf->isref && returnTy != Tvoid) || (irFty.arg_sret != nullptr);
 
-  // Hack around LDC assuming structs and static arrays are in memory:
-  // If the function returns a struct or a static array, and the return
-  // value is not a pointer to a struct or a static array, store it to
-  // a stack slot before continuing.
-  bool storeReturnValueOnStack =
-      (returnTy == Tstruct && !isaPointer(retllval)) ||
-      (returnTy == Tsarray && isaArray(retllval));
-
-  bool retValIsAlloca = false;
-
-  // ignore ABI for intrinsics
-  const bool intrinsic =
-      (dfnval && dfnval->func && DtoIsIntrinsic(dfnval->func));
-  if (!intrinsic && !irFty.arg_sret) {
-    // do ABI specific return value fixups
-    if (storeReturnValueOnStack) {
-      Logger::println("Storing return value to stack slot");
-      LLValue *mem = DtoAlloca(returntype);
-      irFty.getRet(returntype, retllval, mem);
-      retllval = mem;
-      retValIsAlloca = true;
-      storeReturnValueOnStack = false;
+  if (!retValIsLVal) {
+    // let the ABI transform the return value back
+    if (DtoIsInMemoryOnly(returntype)) {
+      retllval = irFty.getRetLVal(returntype, retllval);
+      retValIsLVal = true;
     } else {
-      retllval = irFty.getRet(returntype, retllval);
-      storeReturnValueOnStack =
-          (returnTy == Tstruct && !isaPointer(retllval)) ||
-          (returnTy == Tsarray && isaArray(retllval));
+      retllval = irFty.getRetRVal(returntype, retllval);
     }
   }
 
-  if (storeReturnValueOnStack) {
-    Logger::println("Storing return value to stack slot");
-    retllval = DtoAllocaDump(retllval, returntype);
-    retValIsAlloca = true;
-  }
-
   // repaint the type if necessary
   Type *rbase = stripModifiers(resulttype->toBasetype(), true);
   Type *nextbase = stripModifiers(returntype->toBasetype(), true);
-  bool retinptr = irFty.arg_sret;
   if (!rbase->equals(nextbase)) {
     IF_LOG Logger::println("repainting return value from '%s' to '%s'",
                            returntype->toChars(), rbase->toChars());
@@ -927,7 +945,7 @@ DValue *DtoCallFunction(Loc &loc, Type *resulttype, DValue *fnval,
         LLValue *val =
             DtoInsertValue(llvm::UndefValue::get(DtoType(rbase)), retllval, 0);
         retllval = DtoAllocaDump(val, rbase, ".aalvaluetmp");
-        retinptr = true;
+        retValIsLVal = true;
         break;
       }
     // Fall through.
@@ -974,7 +992,7 @@ DValue *DtoCallFunction(Loc &loc, Type *resulttype, DValue *fnval,
   // set calling convention and parameter attributes
   llvm::AttributeSet &attrlist = attrs;
   if (dfnval && dfnval->func) {
-    LLFunction *llfunc = llvm::dyn_cast<LLFunction>(dfnval->val);
+    LLFunction *llfunc = llvm::dyn_cast<LLFunction>(DtoRVal(dfnval));
     if (llfunc && llfunc->isIntrinsic()) // override intrinsic attrs
     {
       attrlist = llvm::Intrinsic::getAttributes(
@@ -986,6 +1004,10 @@ DValue *DtoCallFunction(Loc &loc, Type *resulttype, DValue *fnval,
   } else {
     call.setCallingConv(callconv);
   }
+  // merge in function attributes set in callOrInvoke
+  attrlist = attrlist.addAttributes(
+      gIR->context(), llvm::AttributeSet::FunctionIndex, call.getAttributes());
+
   call.setAttributes(attrlist);
 
   // Special case for struct constructor calls: For temporaries, using the
@@ -995,14 +1017,11 @@ DValue *DtoCallFunction(Loc &loc, Type *resulttype, DValue *fnval,
   // cleanups (see struct ctor hack in ToElemVisitor::visit(CallExp *)).
   if (dfnval && dfnval->func && dfnval->func->isCtorDeclaration() &&
       dfnval->func->isMember2()->isStructDeclaration()) {
-    return new DVarValue(resulttype, dfnval->vthis);
+    return new DLValue(resulttype, dfnval->vthis);
   }
 
-  // if we are returning through a pointer arg
-  // or if we are returning a reference
-  // make sure we provide a lvalue back!
-  if (retinptr || (tf->isref && returnTy != Tvoid) || retValIsAlloca) {
-    return new DVarValue(resulttype, retllval);
+  if (retValIsLVal) {
+    return new DLValue(resulttype, retllval);
   }
 
   return new DImValue(resulttype, retllval);
