@@ -16,7 +16,7 @@
 #include "gen/logger.h"
 #include "gen/structs.h"
 #include "gen/tollvm.h"
-#include "gen/typeinf.h"
+#include "gen/typinf.h"
 #include "ir/irfunction.h"
 #include "ir/irtypeclass.h"
 #include "ir/irtypestruct.h"
@@ -78,14 +78,14 @@ public:
     }
 
     VarDeclaration *vd = e->var->isVarDeclaration();
-    if (vd && vd->isConst() && vd->init) {
+    if (vd && vd->isConst() && vd->_init) {
       if (vd->inuse) {
         e->error("recursive reference %s", e->toChars());
         result = llvm::UndefValue::get(DtoType(e->type));
       } else {
         vd->inuse++;
         // return the initializer
-        result = DtoConstInitializer(e->loc, e->type, vd->init);
+        result = DtoConstInitializer(e->loc, e->type, vd->_init);
         vd->inuse--;
       }
     }
@@ -158,42 +158,17 @@ public:
                          e->type->toChars());
     LOG_SCOPE;
 
-    Type *t = e->type->toBasetype();
-    Type *cty = t->nextOf()->toBasetype();
+    Type * const t = e->type->toBasetype();
+    Type * const cty = t->nextOf()->toBasetype();
 
-    bool nullterm = (t->ty != Tsarray);
-    size_t endlen = nullterm ? e->len + 1 : e->len;
-
-    LLType *ct = DtoMemType(cty);
-    LLArrayType *at = LLArrayType::get(ct, endlen);
-
-    llvm::StringMap<llvm::GlobalVariable *> *stringLiteralCache = nullptr;
-    LLConstant *_init;
-    switch (cty->size()) {
-    default:
-      llvm_unreachable("Unknown char type");
-    case 1:
-      _init = toConstantArray(ct, at, static_cast<uint8_t *>(e->string), e->len,
-                              nullterm);
-      stringLiteralCache = &(gIR->stringLiteral1ByteCache);
-      break;
-    case 2:
-      _init = toConstantArray(ct, at, static_cast<uint16_t *>(e->string),
-                              e->len, nullterm);
-      stringLiteralCache = &(gIR->stringLiteral2ByteCache);
-      break;
-    case 4:
-      _init = toConstantArray(ct, at, static_cast<uint32_t *>(e->string),
-                              e->len, nullterm);
-      stringLiteralCache = &(gIR->stringLiteral4ByteCache);
-      break;
-    }
+    auto _init = buildStringLiteralConstant(e, t->ty != Tsarray);
 
     if (t->ty == Tsarray) {
       result = _init;
       return;
     }
 
+    auto stringLiteralCache = stringLiteralCacheForType(cty);
     llvm::StringRef key(e->toChars());
     llvm::GlobalVariable *gvar =
         (stringLiteralCache->find(key) == stringLiteralCache->end())
@@ -204,7 +179,11 @@ public:
           llvm::GlobalValue::PrivateLinkage;
       gvar = new llvm::GlobalVariable(gIR->module, _init->getType(), true,
                                       _linkage, _init, ".str");
+#if LDC_LLVM_VER >= 309
+      gvar->setUnnamedAddr(llvm::GlobalValue::UnnamedAddr::Global);
+#else
       gvar->setUnnamedAddr(true);
+#endif
       (*stringLiteralCache)[key] = gvar;
     }
 
@@ -221,7 +200,7 @@ public:
     if (t->ty == Tpointer) {
       result = arrptr;
     } else if (t->ty == Tarray) {
-      LLConstant *clen = LLConstantInt::get(DtoSize_t(), e->len, false);
+      LLConstant *clen = LLConstantInt::get(DtoSize_t(), e->numberOfCodeUnits(), false);
       result = DtoConstSlice(clen, arrptr, e->type);
     } else {
       llvm_unreachable("Unknown type for StringExp.");
@@ -568,7 +547,11 @@ public:
     auto gvar = new llvm::GlobalVariable(
         gIR->module, initval->getType(), canBeConst,
         llvm::GlobalValue::InternalLinkage, initval, ".dynarrayStorage");
+#if LDC_LLVM_VER >= 309
+    gvar->setUnnamedAddr(canBeConst ? llvm::GlobalValue::UnnamedAddr::Global : llvm::GlobalValue::UnnamedAddr::None);
+#else
     gvar->setUnnamedAddr(canBeConst);
+#endif
     llvm::Constant *store = DtoBitCast(gvar, getPtrToType(arrtype));
 
     if (bt->ty == Tpointer) {
@@ -599,15 +582,9 @@ public:
                          e->toChars(), e->type ? e->type->toChars() : "(null)");
     LOG_SCOPE;
 
-    if (e->sinit) {
-      // Copied from VarExp::toConstElem, need to clean this mess up.
-      Type *sdecltype = e->sinit->type->toBasetype();
-      IF_LOG Logger::print("Sym: type=%s\n", sdecltype->toChars());
-      assert(sdecltype->ty == Tstruct);
-      TypeStruct *ts = static_cast<TypeStruct *>(sdecltype);
-      DtoResolveStruct(ts->sym);
-
-      result = getIrAggr(ts->sym)->getDefaultInit();
+    if (e->useStaticInit) {
+      DtoResolveStruct(e->sd);
+      result = getIrAggr(e->sd)->getDefaultInit();
     } else {
       // make sure the struct is resolved
       DtoResolveStruct(e->sd);
@@ -751,7 +728,13 @@ public:
     }
 
     TypeInfoDeclaration *tid = getOrCreateTypeInfoDeclaration(t, nullptr);
-    TypeInfoDeclaration_codegen(tid, p);
+    if ((t->ty == Tclass) &&
+        !static_cast<TypeClass *>(t)->sym->isInterfaceDeclaration()) {
+      // For classes, delegate to special function:
+      TypeInfoClassDeclaration_codegen(tid, p);
+    } else {
+      TypeInfoDeclaration_codegen(tid, p);
+    }
     result = llvm::cast<llvm::GlobalVariable>(getIrGlobal(tid)->value);
   }
 
