@@ -28,7 +28,8 @@
 #include "gen/pragma.h"
 #include "gen/runtime.h"
 #include "gen/structs.h"
-#include "gen/typeinf.h"
+#include "gen/typinf.h"
+#include "gen/uda.h"
 #include "ir/irtype.h"
 #include "ir/irtypeclass.h"
 #include "ir/irtypefunction.h"
@@ -175,6 +176,12 @@ LLType *DtoType(Type *t) {
   case Tenum: {
     Type *bt = t->toBasetype();
     assert(bt);
+    if (t == bt) {
+      // This is an enum forward reference that is only legal when referenced
+      // through an indirection (e.g. "enum E; void foo(E* p);"). For lack of a
+      // better choice, make the outer indirection a void pointer.
+      return getVoidPtrType()->getContainedType(0);
+    }
     return DtoType(bt);
   }
 
@@ -214,26 +221,17 @@ LLType *i1ToI8(LLType *t) {
 
 LLValue *DtoDelegateEquals(TOK op, LLValue *lhs, LLValue *rhs) {
   Logger::println("Doing delegate equality");
-  llvm::Value *b1, *b2;
   if (rhs == nullptr) {
     rhs = LLConstant::getNullValue(lhs->getType());
   }
 
-  LLValue *l = gIR->ir->CreateExtractValue(lhs, 0);
-  LLValue *r = gIR->ir->CreateExtractValue(rhs, 0);
-  b1 = gIR->ir->CreateICmp(llvm::ICmpInst::ICMP_EQ, l, r);
+  LLValue *l1 = gIR->ir->CreateExtractValue(lhs, 0);
+  LLValue *l2 = gIR->ir->CreateExtractValue(lhs, 1);
 
-  l = gIR->ir->CreateExtractValue(lhs, 1);
-  r = gIR->ir->CreateExtractValue(rhs, 1);
-  b2 = gIR->ir->CreateICmp(llvm::ICmpInst::ICMP_EQ, l, r);
+  LLValue *r1 = gIR->ir->CreateExtractValue(rhs, 0);
+  LLValue *r2 = gIR->ir->CreateExtractValue(rhs, 1);
 
-  LLValue *b = gIR->ir->CreateAnd(b1, b2);
-
-  if (op == TOKnotequal || op == TOKnotidentity) {
-    return gIR->ir->CreateNot(b);
-  }
-
-  return b;
+  return createIPairCmp(op, l1, l2, r1, r2);
 }
 
 ////////////////////////////////////////////////////////////////////////////////
@@ -241,12 +239,18 @@ LLValue *DtoDelegateEquals(TOK op, LLValue *lhs, LLValue *rhs) {
 LinkageWithCOMDAT DtoLinkage(Dsymbol *sym) {
   auto linkage = (DtoIsTemplateInstance(sym) ? templateLinkage
                                              : LLGlobalValue::ExternalLinkage);
+
+  // If @(ldc.attributes.weak) is applied, override the linkage to WeakAny
+  if (hasWeakUDA(sym)) {
+    linkage = LLGlobalValue::WeakAnyLinkage;
+  }
+
   return {linkage, supportsCOMDAT()};
 }
 
 bool supportsCOMDAT() {
 #if LDC_LLVM_VER >= 307
-  return !global.params.targetTriple.isOSBinFormatMachO();
+  return !global.params.targetTriple->isOSBinFormatMachO();
 #else
   return false;
 #endif
@@ -458,7 +462,11 @@ LLConstant *DtoConstString(const char *str) {
     gvar = new llvm::GlobalVariable(gIR->module, init->getType(), true,
                                     llvm::GlobalValue::PrivateLinkage, init,
                                     ".str");
+#if LDC_LLVM_VER >= 309
+    gvar->setUnnamedAddr(llvm::GlobalValue::UnnamedAddr::Global);
+#else
     gvar->setUnnamedAddr(true);
+#endif
     gIR->stringLiteral1ByteCache[s] = gvar;
   }
   LLConstant *idxs[] = {DtoConstUint(0), DtoConstUint(0)};
@@ -662,7 +670,7 @@ LLStructType *DtoMutexType() {
   // druntime/src/rt/critical.c
 
   // Windows
-  if (global.params.targetTriple.isOSWindows()) {
+  if (global.params.targetTriple->isOSWindows()) {
     llvm::Type *VoidPtrTy = llvm::Type::getInt8PtrTy(gIR->context());
     llvm::Type *Int32Ty = llvm::Type::getInt32Ty(gIR->context());
 
@@ -691,15 +699,15 @@ LLStructType *DtoMutexType() {
   }
 
   // FreeBSD, NetBSD, OpenBSD, DragonFly
-  if (global.params.targetTriple.isOSFreeBSD() ||
+  if (global.params.targetTriple->isOSFreeBSD() ||
 #if LDC_LLVM_VER > 305
-      global.params.targetTriple.isOSNetBSD() ||
-      global.params.targetTriple.isOSOpenBSD() ||
-      global.params.targetTriple.isOSDragonFly()
+      global.params.targetTriple->isOSNetBSD() ||
+      global.params.targetTriple->isOSOpenBSD() ||
+      global.params.targetTriple->isOSDragonFly()
 #else
-      global.params.targetTriple.getOS() == llvm::Triple::NetBSD ||
-      global.params.targetTriple.getOS() == llvm::Triple::OpenBSD ||
-      global.params.targetTriple.getOS() == llvm::Triple::DragonFly
+      global.params.targetTriple->getOS() == llvm::Triple::NetBSD ||
+      global.params.targetTriple->getOS() == llvm::Triple::OpenBSD ||
+      global.params.targetTriple->getOS() == llvm::Triple::DragonFly
 #endif
           ) {
     // Just a pointer
