@@ -20,6 +20,8 @@
 // Copyright (c) 2007 by Tomas Lindquist Olsen
 // tomas at famolsen dk
 
+#include "gen/typinf.h"
+
 #include "aggregate.h"
 #include "attrib.h"
 #include "declaration.h"
@@ -102,13 +104,13 @@ TypeInfoDeclaration *getOrCreateTypeInfoDeclaration(Type *torig, Scope *sc) {
   Type *t = torig->merge2(); // do this since not all Type's are merge'd
   if (!t->vtinfo) {
     if (t->isShared()) { // does both 'shared' and 'shared const'
-      t->vtinfo = new TypeInfoSharedDeclaration(t);
+      t->vtinfo = TypeInfoSharedDeclaration::create(t);
     } else if (t->isConst()) {
-      t->vtinfo = new TypeInfoConstDeclaration(t);
+      t->vtinfo = TypeInfoConstDeclaration::create(t);
     } else if (t->isImmutable()) {
-      t->vtinfo = new TypeInfoInvariantDeclaration(t);
+      t->vtinfo = TypeInfoInvariantDeclaration::create(t);
     } else if (t->isWild()) {
-      t->vtinfo = new TypeInfoWildDeclaration(t);
+      t->vtinfo = TypeInfoWildDeclaration::create(t);
     } else {
       t->vtinfo = createUnqualified(t);
     }
@@ -217,10 +219,10 @@ static void emitTypeMetadata(TypeInfoDeclaration *tid) {
 }
 
 void DtoResolveTypeInfo(TypeInfoDeclaration *tid) {
-  if (tid->ir.isResolved()) {
+  if (tid->ir->isResolved()) {
     return;
   }
-  tid->ir.setResolved();
+  tid->ir->setResolved();
 
   // TypeInfo instances (except ClassInfo ones) are always emitted as weak
   // symbols when they are used. We call semanticTypeInfo() to make sure
@@ -271,30 +273,15 @@ public:
     b.push_string(sd->toPrettyChars());
 
     // void[] init
-    // emit void[] with the default initialier, the array is null if the default
-    // initializer is zero
+    // the array is null if the default initializer is zero
     if (!sd->members || decl->tinfo->isZeroInit(decl->loc)) {
       b.push_null_void_array();
     }
     // otherwise emit a void[] with the default initializer
     else {
-      Type *memtype = sd->memtype;
-      LLType *memty = DtoType(memtype);
-      LLConstant *C;
       Expression *defaultval = sd->getDefaultValue(decl->loc);
-      if (memtype->isintegral()) {
-        C = LLConstantInt::get(memty, defaultval->toInteger(),
-                               !isLLVMUnsigned(memtype));
-      } else if (memtype->isString()) {
-        C = DtoConstString(
-            static_cast<const char *>(defaultval->toStringExp()->string));
-      } else if (memtype->isfloating()) {
-        C = LLConstantFP::get(memty, defaultval->toReal());
-      } else {
-        llvm_unreachable("Unsupported type");
-      }
-
-      b.push_void_array(C, memtype, sd);
+      LLConstant *c = toConstElem(defaultval, gIR);
+      b.push_void_array(c, sd->memtype, sd);
     }
 
     // finish
@@ -469,7 +456,7 @@ public:
     // (m_arg1/m_arg2) which are used for the X86_64 System V ABI varargs
     // implementation. They are not present on any other cpu/os.
     unsigned expectedFields = 12;
-    if (global.params.targetTriple.getArch() == llvm::Triple::x86_64) {
+    if (global.params.targetTriple->getArch() == llvm::Triple::x86_64) {
       expectedFields += 2;
     }
     if (Type::typeinfostruct->fields.dim != expectedFields) {
@@ -697,14 +684,14 @@ public:
 /* ========================================================================= */
 
 void TypeInfoDeclaration_codegen(TypeInfoDeclaration *decl, IRState *p) {
-  IF_LOG Logger::println("TypeInfoDeclaration::codegen(%s)",
+  IF_LOG Logger::println("TypeInfoDeclaration_codegen(%s)",
                          decl->toPrettyChars());
   LOG_SCOPE;
 
-  if (decl->ir.isDefined()) {
+  if (decl->ir->isDefined()) {
     return;
   }
-  decl->ir.setDefined();
+  decl->ir->setDefined();
 
   std::string mangled(mangle(decl));
   IF_LOG {
@@ -726,8 +713,11 @@ void TypeInfoDeclaration_codegen(TypeInfoDeclaration *decl, IRState *p) {
     } else {
       irg->type = LLStructType::create(gIR->context(), decl->toPrettyChars());
     }
-    LLGlobalVariable *g = new LLGlobalVariable(gIR->module, irg->type, true,
-                                               lwc.first, nullptr, mangled);
+    // Create the symbol. We need to keep it mutable as the type is not declared
+    // as immutable on the D side, and e.g. synchronized() can be used on the
+    // implicit monitor.
+    auto g = new LLGlobalVariable(gIR->module, irg->type, false, lwc.first,
+                                  nullptr, mangled);
     setLinkage(lwc, g);
     irg->value = g;
   }
@@ -747,6 +737,10 @@ void TypeInfoDeclaration_codegen(TypeInfoDeclaration *decl, IRState *p) {
 /* ========================================================================= */
 
 void TypeInfoClassDeclaration_codegen(TypeInfoDeclaration *decl, IRState *p) {
+  IF_LOG Logger::println("TypeInfoClassDeclaration_codegen(%s)",
+                         decl->toPrettyChars());
+  LOG_SCOPE;
+
   // For classes, the TypeInfo is in fact a ClassInfo instance and emitted
   // as a __ClassZ symbol. For interfaces, the __InterfaceZ symbol is
   // referenced as "info" member in a (normal) TypeInfo_Interface instance.
