@@ -29,6 +29,7 @@
 #include "gen/runtime.h"
 #include "gen/functions.h"
 #include "gen/abi.h"
+#include "gen/mangling.h"
 
 #include "ir/iraggr.h"
 #include "ir/irfunction.h"
@@ -45,10 +46,8 @@ LLGlobalVariable *IrAggr::getVtblSymbol() {
     return vtbl;
   }
 
-  // create the initZ symbol
-  std::string initname("_D");
-  initname.append(mangle(aggrdecl));
-  initname.append("6__vtblZ");
+  // create the vtblZ symbol
+  auto initname = getMangledVTableSymbolName(aggrdecl);
 
   LLType *vtblTy = stripModifiers(type)->ctype->isClass()->getVtbl();
 
@@ -66,15 +65,8 @@ LLGlobalVariable *IrAggr::getClassInfoSymbol() {
     return classInfo;
   }
 
-  // create the initZ symbol
-  std::string initname("_D");
-  initname.append(mangle(aggrdecl));
-
-  if (aggrdecl->isInterfaceDeclaration()) {
-    initname.append("11__InterfaceZ");
-  } else {
-    initname.append("7__ClassZ");
-  }
+  // create the ClassZ / InterfaceZ symbol
+  std::string initname = getMangledClassInfoSymbolName(aggrdecl);
 
   // The type is also ClassInfo for interfaces – the actual TypeInfo for them
   // is a TypeInfo_Interface instance that references __ClassZ in its "base"
@@ -196,16 +188,14 @@ LLConstant *IrAggr::getVtblInit() {
       DtoResolveFunction(fd);
       assert(isIrFuncCreated(fd) && "invalid vtbl function");
       c = getIrFunc(fd)->func;
-      if (cd->isFuncHidden(
-              fd)) { /* fd is hidden from the view of this class.
-                      * If fd overlaps with any function in the vtbl[], then
-                      * issue 'hidden' error.
-                      */
+      if (cd->isFuncHidden(fd)) {
+        // fd is hidden from the view of this class. If fd overlaps with any
+        // function in the vtbl[], issue error.
         for (size_t j = 1; j < n; j++) {
           if (j == i) {
             continue;
           }
-          FuncDeclaration *fd2 =
+          auto fd2 =
               static_cast<Dsymbol *>(cd->vtbl.data[j])->isFuncDeclaration();
           if (!fd2->ident->equals(fd->ident)) {
             continue;
@@ -272,14 +262,14 @@ LLConstant *IrAggr::getClassInfoInit() {
 
 llvm::GlobalVariable *IrAggr::getInterfaceVtbl(BaseClass *b, bool new_instance,
                                                size_t interfaces_index) {
-  auto it = interfaceVtblMap.find(b->base);
+  auto it = interfaceVtblMap.find({b->sym, interfaces_index});
   if (it != interfaceVtblMap.end()) {
     return it->second;
   }
 
   IF_LOG Logger::println(
       "Building vtbl for implementation of interface %s in class %s",
-      b->base->toPrettyChars(), aggrdecl->toPrettyChars());
+      b->sym->toPrettyChars(), aggrdecl->toPrettyChars());
   LOG_SCOPE;
 
   ClassDeclaration *cd = aggrdecl->isClassDeclaration();
@@ -291,7 +281,7 @@ llvm::GlobalVariable *IrAggr::getInterfaceVtbl(BaseClass *b, bool new_instance,
   std::vector<llvm::Constant *> constants;
   constants.reserve(vtbl_array.dim);
 
-  if (!b->base->isCPPinterface()) { // skip interface info for CPP interfaces
+  if (!b->sym->isCPPinterface()) { // skip interface info for CPP interfaces
     // index into the interfaces array
     llvm::Constant *idxs[2] = {DtoConstSize_t(0),
                                DtoConstSize_t(interfaces_index)};
@@ -306,9 +296,15 @@ llvm::GlobalVariable *IrAggr::getInterfaceVtbl(BaseClass *b, bool new_instance,
     constants.push_back(c);
   }
 
+  // Thunk prefix
+  char thunkPrefix[16];
+  int thunkLen = sprintf(thunkPrefix, "Th%d", b->offset);
+  char thunkPrefixLen[16];
+  sprintf(thunkPrefixLen, "%d", thunkLen);
+
   // add virtual function pointers
   size_t n = vtbl_array.dim;
-  for (size_t i = b->base->vtblOffset(); i < n; i++) {
+  for (size_t i = b->sym->vtblOffset(); i < n; i++) {
     Dsymbol *dsym = static_cast<Dsymbol *>(vtbl_array.data[i]);
     if (dsym == nullptr) {
       // FIXME
@@ -331,11 +327,18 @@ llvm::GlobalVariable *IrAggr::getInterfaceVtbl(BaseClass *b, bool new_instance,
 
     assert(irFunc->irFty.arg_this);
 
+    int thunkOffset = b->offset;
+    if (fd->interfaceVirtual)
+      thunkOffset -= fd->interfaceVirtual->offset;
+    if (thunkOffset == 0) {
+      constants.push_back(irFunc->func);
+      continue;
+    }
+
     // Create the thunk function if it does not already exist in this
     // module.
     OutBuffer nameBuf;
-    nameBuf.writestring("Th");
-    nameBuf.printf("%i", b->offset);
+    nameBuf.writestring(thunkPrefix);
     nameBuf.writestring(mangleExact(fd));
     const char *thunkName = nameBuf.extractString();
     llvm::Function *thunk = gIR->module.getFunction(thunkName);
@@ -350,13 +353,42 @@ llvm::GlobalVariable *IrAggr::getInterfaceVtbl(BaseClass *b, bool new_instance,
 
       // Thunks themselves don't have an identity, only the target
       // function has.
+#if LDC_LLVM_VER >= 309
+      thunk->setUnnamedAddr(llvm::GlobalValue::UnnamedAddr::Global);
+#else
       thunk->setUnnamedAddr(true);
+#endif
+
+#if LDC_LLVM_VER >= 307
+      // thunks don't need exception handling themselves
+      thunk->setPersonalityFn(nullptr);
+#endif
+
+      // It is necessary to add debug information to the thunk in case it is
+      // subject to inlining. See https://llvm.org/bugs/show_bug.cgi?id=26833
+      IF_LOG Logger::println("Doing function body for thunk to: %s",
+                             fd->toChars());
+
+      // Create a dummy FuncDeclaration with enough information to satisfy the
+      // DIBuilder
+      FuncDeclaration *thunkFd = reinterpret_cast<FuncDeclaration *>(memcpy(
+          new char[sizeof(FuncDeclaration)], fd, sizeof(FuncDeclaration)));
+      thunkFd->ir = new IrDsymbol();
+      auto thunkFunc = getIrFunc(thunkFd, true); // create the IrFunction
+      thunkFunc->func = thunk;
+      thunkFunc->type = irFunc->type;
+      gIR->functions.push_back(thunkFunc);
+
+      // debug info
+      thunkFunc->diSubprogram = gIR->DBuilder.EmitThunk(thunk, thunkFd);
 
       // create entry and end blocks
       llvm::BasicBlock *beginbb =
           llvm::BasicBlock::Create(gIR->context(), "", thunk);
       gIR->scopes.push_back(IRScope(beginbb));
 
+      gIR->DBuilder.EmitFuncStart(thunkFd);
+
       // Copy the function parameters, so later we can pass them to the
       // real function and set their names from the original function (the
       // latter being just for IR readablilty).
@@ -369,15 +401,20 @@ llvm::GlobalVariable *IrAggr::getInterfaceVtbl(BaseClass *b, bool new_instance,
       }
 
       // cast 'this' to Object
-      LLValue *&thisArg = args[(!irFunc->irFty.arg_sret ||
-                                gABI->passThisBeforeSret(irFunc->type))
-                                   ? 0
-                                   : 1];
+      const int thisArgIndex =
+          (!irFunc->irFty.arg_sret || gABI->passThisBeforeSret(irFunc->type))
+              ? 0
+              : 1;
+      LLValue *&thisArg = args[thisArgIndex];
       LLType *targetThisType = thisArg->getType();
       thisArg = DtoBitCast(thisArg, getVoidPtrType());
-      thisArg = DtoGEP1(thisArg, DtoConstInt(-b->offset), true);
+      thisArg = DtoGEP1(thisArg, DtoConstInt(-thunkOffset), true);
       thisArg = DtoBitCast(thisArg, targetThisType);
 
+      // all calls that might be subject to inlining into a caller with debug
+      // info should have debug info, too
+      gIR->DBuilder.EmitStopPoint(fd->loc);
+
       // call the real vtbl function.
       llvm::CallSite call = gIR->ir->CreateCall(irFunc->func, args);
       call.setCallingConv(irFunc->func->getCallingConv());
@@ -390,8 +427,12 @@ llvm::GlobalVariable *IrAggr::getInterfaceVtbl(BaseClass *b, bool new_instance,
                                  beginbb);
       }
 
+      gIR->DBuilder.EmitFuncEnd(thunkFd);
+
       // clean up
       gIR->scopes.pop_back();
+
+      gIR->functions.pop_back();
     }
 
     constants.push_back(thunk);
@@ -404,7 +445,9 @@ llvm::GlobalVariable *IrAggr::getInterfaceVtbl(BaseClass *b, bool new_instance,
   std::string mangledName("_D");
   mangledName.append(mangle(cd));
   mangledName.append("11__interface");
-  mangledName.append(mangle(b->base));
+  mangledName.append(mangle(b->sym));
+  mangledName.append(thunkPrefixLen);
+  mangledName.append(thunkPrefix);
   mangledName.append("6__vtblZ");
 
   const auto lwc = DtoLinkage(cd);
@@ -414,7 +457,7 @@ llvm::GlobalVariable *IrAggr::getInterfaceVtbl(BaseClass *b, bool new_instance,
   setLinkage(lwc, GV);
 
   // insert into the vtbl map
-  interfaceVtblMap.insert(std::make_pair(b->base, GV));
+  interfaceVtblMap.insert({{b->sym, interfaces_index}, GV});
 
   return GV;
 }
@@ -463,9 +506,9 @@ LLConstant *IrAggr::getClassInfoInterfaces() {
   for (size_t i = 0; i < n; ++i) {
     BaseClass *it = interfacesWithVtbls[i];
 
-    IF_LOG Logger::println("Adding interface %s", it->base->toPrettyChars());
+    IF_LOG Logger::println("Adding interface %s", it->sym->toPrettyChars());
 
-    IrAggr *irinter = getIrAggr(it->base);
+    IrAggr *irinter = getIrAggr(it->sym);
     assert(irinter && "interface has null IrStruct");
     IrTypeClass *itc = stripModifiers(irinter->type)->ctype->isClass();
     assert(itc && "null interface IrTypeClass");
@@ -480,7 +523,7 @@ LLConstant *IrAggr::getClassInfoInterfaces() {
     if (cd->isInterfaceDeclaration()) {
       vtb = DtoConstSlice(DtoConstSize_t(0), getNullValue(voidptrptr_type));
     } else {
-      auto itv = interfaceVtblMap.find(it->base);
+      auto itv = interfaceVtblMap.find({it->sym, i});
       assert(itv != interfaceVtblMap.end() && "interface vtbl not found");
       vtb = itv->second;
       vtb = DtoBitCast(vtb, voidptrptr_type);
@@ -502,8 +545,9 @@ LLConstant *IrAggr::getClassInfoInterfaces() {
 
   // create and apply initializer
   LLConstant *arr = LLConstantArray::get(array_type, constants);
-  classInterfacesArray->setInitializer(arr);
-  setLinkage(cd, classInterfacesArray);
+  auto ciarr = getInterfaceArraySymbol();
+  ciarr->setInitializer(arr);
+  setLinkage(cd, ciarr);
 
   // return null, only baseclass provide interfaces
   if (cd->vtblInterfaces->dim == 0) {
@@ -517,9 +561,9 @@ LLConstant *IrAggr::getClassInfoInterfaces() {
 
   LLConstant *ptr = llvm::ConstantExpr::getGetElementPtr(
 #if LDC_LLVM_VER >= 307
-      isaPointer(classInterfacesArray)->getElementType(),
+      isaPointer(ciarr)->getElementType(),
 #endif
-      classInterfacesArray, idxs, true);
+      ciarr, idxs, true);
 
   // return as a slice
   return DtoConstSlice(DtoConstSize_t(cd->vtblInterfaces->dim), ptr);
