@@ -19,6 +19,7 @@
 #include "llvm/ADT/DenseMapInfo.h"
 #include "gen/llvm.h"
 #include "gen/irstate.h"
+#include "gen/pgo.h"
 #include "ir/irfuncty.h"
 #include <map>
 #include <stack>
@@ -98,6 +99,9 @@ struct CleanupExitTarget {
   /// stores to the branch selector variable when converting from one to two
   /// targets.
   std::vector<llvm::BasicBlock *> sourceBlocks;
+
+  /// MSVC: The basic blocks that are executed when going this route
+  std::vector<llvm::BasicBlock *> cleanupBlocks;
 };
 
 /// Represents a scope (in abstract terms, not curly braces) that requires a
@@ -148,6 +152,10 @@ public:
   /// and popped again once it is left. If the corresponding landing pad has
   /// not been generated yet (this is done lazily), the pointer is null.
   std::vector<llvm::BasicBlock *> landingPads;
+
+  /// MSVC: The original basic blocks that are executed for beginBlock to
+  /// endBlock
+  std::vector<llvm::BasicBlock *> cleanupBlocks;
 };
 
 /// Stores information to be able to branch to a catch clause if it matches.
@@ -165,8 +173,12 @@ struct CatchScope {
   /// The cleanup scope stack level corresponding to this catch.
   CleanupCursor cleanupScope;
 
+  // PGO branch weights for the exception type match branch.
+  // (first weight is for match, second is for mismatch)
+  llvm::MDNode *branchWeights = nullptr;
+
   CatchScope(llvm::Constant *classInfoPtr, llvm::BasicBlock *bodyBlock,
-             CleanupCursor cleanupScope);
+             CleanupCursor cleanupScope, llvm::MDNode *branchWeights = nullptr);
 };
 
 /// Keeps track of active (abstract) scopes in a function that influence code
@@ -210,6 +222,13 @@ public:
   /// reached.
   void runAllCleanups(llvm::BasicBlock *continueWith);
 
+#if LDC_LLVM_VER >= 308
+  void runCleanupCopies(CleanupCursor sourceScope, CleanupCursor targetScope,
+                        llvm::BasicBlock *continueWith);
+  llvm::BasicBlock *runCleanupPad(CleanupCursor scope,
+                                  llvm::BasicBlock *unwindTo);
+#endif
+
   /// Pops all the cleanups between the current scope and the target cursor.
   ///
   /// This does not insert any cleanup calls, use #runCleanups() beforehand.
@@ -230,11 +249,34 @@ public:
   /// given ClassInfo constant and to branch to the given body block if it
   /// matches. The registered catch blocks are maintained on a stack, with the
   /// top-most (i.e. last pushed, innermost) taking precedence.
-  void pushCatch(llvm::Constant *classInfoPtr, llvm::BasicBlock *bodyBlock);
+  void pushCatch(llvm::Constant *classInfoPtr, llvm::BasicBlock *bodyBlock,
+                 llvm::MDNode *matchWeights = nullptr);
 
   /// Unregisters the last registered catch block.
   void popCatch();
 
+  size_t currentCatchScope() { return catchScopes.size(); }
+
+#if LDC_LLVM_VER >= 308
+  /// MSVC: catch and cleanup code is emitted as funclets and need
+  /// to be referenced from inner pads and calls
+  void pushFunclet(llvm::Value *funclet) {
+    funclets.push_back(funclet);
+  }
+
+  void popFunclet() {
+    funclets.pop_back();
+  }
+
+  llvm::Value *getFunclet() {
+    return funclets.empty() ? nullptr : funclets.back();
+  }
+  llvm::Value *getFuncletToken() {
+    return funclets.empty() ? llvm::ConstantTokenNone::get(irs->context())
+                            : funclets.back();
+  }
+#endif
+
   /// Registers a loop statement to be used as a target for break/continue
   /// statements in the current scope.
   void pushLoopTarget(Statement *loopStatement,
@@ -295,6 +337,9 @@ public:
   /// the way there.
   void breakToClosest() { jumpToClosest(breakTargets); }
 
+  /// get exisiting or emit new landing pad
+  llvm::BasicBlock *getLandingPad();
+
 private:
   /// Internal version that allows specifying the scope at which to start
   /// emitting the cleanups.
@@ -305,9 +350,15 @@ private:
 
   std::vector<llvm::BasicBlock *> &currentLandingPads();
 
+  llvm::BasicBlock * &getLandingPadRef(CleanupCursor scope);
+
   /// Emits a landing pad to honor all the active cleanups and catches.
   llvm::BasicBlock *emitLandingPad();
 
+#if LDC_LLVM_VER >= 308
+  llvm::BasicBlock *emitLandingPadMSVCEH(CleanupCursor scope);
+#endif
+
   /// Unified implementation for labeled break/continue.
   void jumpToStatement(std::vector<JumpTarget> &targets,
                        Statement *loopOrSwitchStatement);
@@ -347,6 +398,9 @@ private:
   /// (null if not yet emitted, one element is pushed to/popped from the back
   /// on entering/leaving a catch block).
   std::vector<llvm::BasicBlock *> topLevelLandingPads;
+
+  /// MSVC: stack of currently built catch/cleanup funclets
+  std::vector<llvm::Value*> funclets;
 };
 
 template <typename T>
@@ -360,31 +414,35 @@ llvm::CallSite ScopeStack::callOrInvoke(llvm::Value *callee, const T &args,
   const bool doesNotThrow =
       calleeFn && (calleeFn->isIntrinsic() || calleeFn->doesNotThrow());
 
+#if LDC_LLVM_VER >= 308
+  // calls inside a funclet must be annotated with its value
+  llvm::SmallVector<llvm::OperandBundleDef, 2> BundleList;
+  if (auto funclet = getFunclet())
+    BundleList.push_back(llvm::OperandBundleDef("funclet", funclet));
+#endif
+
   if (doesNotThrow || (cleanupScopes.empty() && catchScopes.empty())) {
-    llvm::CallInst *call = irs->ir->CreateCall(callee, args, name);
+    llvm::CallInst *call = irs->ir->CreateCall(callee, args,
+#if LDC_LLVM_VER >= 308
+                                               BundleList, 
+#endif
+                                               name);
     if (calleeFn) {
       call->setAttributes(calleeFn->getAttributes());
     }
     return call;
   }
 
-  if (currentLandingPads().empty()) {
-    // Have not encountered any catches (for which we would push a scope) or
-    // calls to throwing functions (where we would have already executed
-    // this if) in this cleanup scope yet.
-    currentLandingPads().push_back(nullptr);
-  }
-
-  llvm::BasicBlock *&landingPad = currentLandingPads().back();
-  if (!landingPad) {
-    landingPad = emitLandingPad();
-  }
+  llvm::BasicBlock* landingPad = getLandingPad();
 
   llvm::BasicBlock *postinvoke = llvm::BasicBlock::Create(
       irs->context(), "postinvoke", irs->topfunc(), landingPad);
   llvm::InvokeInst *invoke =
-      irs->ir->CreateInvoke(callee, postinvoke, landingPad, args, name);
-
+      irs->ir->CreateInvoke(callee, postinvoke, landingPad, args,
+#if LDC_LLVM_VER >= 308
+                            BundleList, 
+#endif
+                            name);
   if (calleeFn) {
     invoke->setAttributes(calleeFn->getAttributes());
   }
@@ -465,8 +523,15 @@ struct IrFunction {
   // Debug info for all variables
   VariableMap variableMap;
 
+  // PGO information
+  CodeGenPGO pgo;
+
   IrFuncTy irFty;
 
+  /// Stores the FastMath options for this functions.
+  /// These are set e.g. by math related UDA's from ldc.attributes.
+  llvm::FastMathFlags FMF;
+
 private:
   llvm::AllocaInst *ehPtrSlot = nullptr;
   llvm::BasicBlock *resumeUnwindBlock = nullptr;
@@ -474,5 +539,6 @@ private:
 
 IrFunction *getIrFunc(FuncDeclaration *decl, bool create = false);
 bool isIrFuncCreated(FuncDeclaration *decl);
+bool useMSVCEH();
 
 #endif
