@@ -23,8 +23,6 @@
 #include "gen/functions.h"
 #include "ir/irtypeclass.h"
 
-//////////////////////////////////////////////////////////////////////////////
-
 IrTypeClass::IrTypeClass(ClassDeclaration *cd)
     : IrTypeAggr(cd), cd(cd), tc(static_cast<TypeClass *>(cd->type)) {
   std::string vtbl_name(cd->toPrettyChars());
@@ -33,50 +31,47 @@ IrTypeClass::IrTypeClass(ClassDeclaration *cd)
   vtbl_size = cd->vtbl.dim;
 }
 
-//////////////////////////////////////////////////////////////////////////////
-
-void IrTypeClass::addBaseClassData(AggrTypeBuilder &builder,
-                                   ClassDeclaration *base) {
-  if (base->baseClass) {
-    addBaseClassData(builder, base->baseClass);
+void IrTypeClass::addClassData(AggrTypeBuilder &builder,
+                               ClassDeclaration *currCd) {
+  // First, recursively add the fields for our base class and interfaces, if
+  // any.
+  if (currCd->baseClass) {
+    addClassData(builder, currCd->baseClass);
   }
 
-  builder.addAggregate(base);
+  if (currCd->vtblInterfaces && currCd->vtblInterfaces->dim > 0) {
+    // KLUDGE: The first pointer in the vtbl will be of type object.Interface;
+    // extract that from the "well-known" object.TypeInfo_Class definition.
+    // For C++ interfaces, this vtbl entry has to be omitted
+    const auto interfaceArrayType = Type::typeinfoclass->fields[3]->type;
+    const auto interfacePtrType = interfaceArrayType->nextOf()->pointerTo();
 
-  // any interface implementations?
-  if (base->vtblInterfaces && base->vtblInterfaces->dim > 0) {
-    bool new_instances = (base == cd);
-
-    VarDeclaration *interfaces_idx = Type::typeinfoclass->fields[3];
-    Type *first = interfaces_idx->type->nextOf()->pointerTo();
-
-    // align offset
     builder.alignCurrentOffset(Target::ptrsize);
 
-    for (auto b : *base->vtblInterfaces) {
+    for (auto b : *currCd->vtblInterfaces) {
       IF_LOG Logger::println("Adding interface vtbl for %s",
-                             b->base->toPrettyChars());
+                             b->sym->toPrettyChars());
 
       FuncDeclarations arr;
-      b->fillVtbl(cd, &arr, new_instances);
+      b->fillVtbl(cd, &arr, currCd == cd);
 
       // add to the interface map
-      addInterfaceToMap(b->base, builder.currentFieldIndex());
-
-      llvm::Type *ivtbl_type =
+      addInterfaceToMap(b->sym, builder.currentFieldIndex());
+      Type* first = b->sym->isCPPinterface() ? nullptr : interfacePtrType;
+      const auto ivtblType =
           llvm::StructType::get(gIR->context(), buildVtblType(first, &arr));
-      builder.addType(llvm::PointerType::get(ivtbl_type, 0), Target::ptrsize);
+      builder.addType(llvm::PointerType::get(ivtblType, 0), Target::ptrsize);
 
-      // inc count
-      num_interface_vtbls++;
+      ++num_interface_vtbls;
     }
   }
+
+  // Finally, the data members for this class.
+  builder.addAggregate(currCd);
 }
 
-//////////////////////////////////////////////////////////////////////////////
-
 IrTypeClass *IrTypeClass::get(ClassDeclaration *cd) {
-  auto t = new IrTypeClass(cd);
+  const auto t = new IrTypeClass(cd);
   cd->type->ctype = t;
 
   IF_LOG Logger::println("Building class type %s @ %s", cd->toPrettyChars(),
@@ -84,10 +79,9 @@ IrTypeClass *IrTypeClass::get(ClassDeclaration *cd) {
   LOG_SCOPE;
   IF_LOG Logger::println("Instance size: %u", cd->structsize);
 
-  // This class may contain an align declaration. See issue 726.
+  // This class may contain an align declaration. See GitHub #726.
   t->packed = false;
-  for (ClassDeclaration *base = cd; base != nullptr && !t->packed;
-       base = base->baseClass) {
+  for (auto base = cd; base != nullptr && !t->packed; base = base->baseClass) {
     t->packed = isPacked(base);
   }
 
@@ -96,12 +90,11 @@ IrTypeClass *IrTypeClass::get(ClassDeclaration *cd) {
   // add vtbl
   builder.addType(llvm::PointerType::get(t->vtbl_type, 0), Target::ptrsize);
 
-  // interfaces are just a vtable
   if (cd->isInterfaceDeclaration()) {
+    // interfaces are just a vtable
     t->num_interface_vtbls = cd->vtblInterfaces ? cd->vtblInterfaces->dim : 0;
-  }
-  // classes have monitor and fields
-  else {
+  } else {
+    // classes have monitor and fields
     if (!cd->isCPPclass() && !cd->isCPPinterface()) {
       // add monitor
       builder.addType(
@@ -110,13 +103,12 @@ IrTypeClass *IrTypeClass::get(ClassDeclaration *cd) {
     }
 
     // add data members recursively
-    t->addBaseClassData(builder, cd);
+    t->addClassData(builder, cd);
 
     // add tail padding
     builder.addTailPadding(cd->structsize);
   }
 
-  // errors are fatal during codegen
   if (global.errors) {
     fatal();
   }
@@ -125,26 +117,24 @@ IrTypeClass *IrTypeClass::get(ClassDeclaration *cd) {
   isaStruct(t->type)->setBody(builder.defaultTypes(), t->packed);
   t->varGEPIndices = builder.varGEPIndices();
 
-  // VTBL
-
   // set vtbl type body
   FuncDeclarations vtbl;
   vtbl.reserve(cd->vtbl.dim);
-  vtbl.push(nullptr);
+  if (!cd->isCPPclass())
+    vtbl.push(nullptr);
   for (size_t i = cd->vtblOffset(); i < cd->vtbl.dim; ++i) {
     FuncDeclaration *fd = cd->vtbl[i]->isFuncDeclaration();
     assert(fd);
     vtbl.push(fd);
   }
-  t->vtbl_type->setBody(t->buildVtblType(Type::typeinfoclass->type, &vtbl));
+  Type* first = cd->isCPPclass() ? nullptr : Type::typeinfoclass->type;
+  t->vtbl_type->setBody(t->buildVtblType(first, &vtbl));
 
   IF_LOG Logger::cout() << "class type: " << *t->type << std::endl;
 
   return t;
 }
 
-//////////////////////////////////////////////////////////////////////////////
-
 std::vector<llvm::Type *>
 IrTypeClass::buildVtblType(Type *first, FuncDeclarations *vtbl_array) {
   IF_LOG Logger::println("Building vtbl type for class %s",
@@ -154,18 +144,18 @@ IrTypeClass::buildVtblType(Type *first, FuncDeclarations *vtbl_array) {
   std::vector<llvm::Type *> types;
   types.reserve(vtbl_array->dim);
 
-  // first comes the classinfo
-  if (!cd->isCPPclass() && !cd->isCPPinterface()) {
+  auto I = vtbl_array->begin();
+  // first comes the classinfo for D interfaces
+  if (first) {
     types.push_back(DtoType(first));
+    ++I;
   }
 
   // then come the functions
-  for (auto I = vtbl_array->begin() + 1, E = vtbl_array->end(); I != E; ++I) {
+  for (auto E = vtbl_array->end(); I != E; ++I) {
     FuncDeclaration *fd = *I;
     if (fd == nullptr) {
-      // FIXME
-      // why is this null?
-      // happens for mini/s.d
+      // FIXME: This stems from the ancient D1 days – can it still happen?
       types.push_back(getVoidPtrType());
       continue;
     }
@@ -179,7 +169,7 @@ IrTypeClass::buildVtblType(Type *first, FuncDeclarations *vtbl_array) {
       Logger::println("Running late semantic3 to infer return type.");
       TemplateInstance *spec = fd->isSpeculative();
       unsigned int olderrs = global.errors;
-      fd->semantic3(fd->scope);
+      fd->semantic3(fd->_scope);
       if (spec && global.errors != olderrs) {
         spec->errors = global.errors - olderrs;
       }
@@ -200,16 +190,10 @@ IrTypeClass::buildVtblType(Type *first, FuncDeclarations *vtbl_array) {
   return types;
 }
 
-//////////////////////////////////////////////////////////////////////////////
-
 llvm::Type *IrTypeClass::getLLType() { return llvm::PointerType::get(type, 0); }
 
-//////////////////////////////////////////////////////////////////////////////
-
 llvm::Type *IrTypeClass::getMemoryLLType() { return type; }
 
-//////////////////////////////////////////////////////////////////////////////
-
 size_t IrTypeClass::getInterfaceIndex(ClassDeclaration *inter) {
   auto it = interfaceMap.find(inter);
   if (it == interfaceMap.end()) {
@@ -218,8 +202,6 @@ size_t IrTypeClass::getInterfaceIndex(ClassDeclaration *inter) {
   return it->second;
 }
 
-//////////////////////////////////////////////////////////////////////////////
-
 void IrTypeClass::addInterfaceToMap(ClassDeclaration *inter, size_t index) {
   // don't duplicate work or overwrite indices
   if (interfaceMap.find(inter) != interfaceMap.end()) {
@@ -231,10 +213,8 @@ void IrTypeClass::addInterfaceToMap(ClassDeclaration *inter, size_t index) {
 
   // add the direct base interfaces recursively - they
   // are accessed through the same index
-  if (inter->interfaces_dim > 0) {
-    BaseClass *b = inter->interfaces[0];
-    addInterfaceToMap(b->base, index);
+  if (inter->interfaces.length > 0) {
+    BaseClass *b = inter->interfaces.ptr[0];
+    addInterfaceToMap(b->sym, index);
   }
 }
-
-//////////////////////////////////////////////////////////////////////////////
